/*=======================================================================================================*/
/*      Distributed Speech Recognition                                                                   */
/*      ETSI ES 202 050 Advanced Front-End Feature Extraction Algorithm & Compression Algorithm &        */
/*      ETSI ES 202 212 Extended Advanced Front-End Feature Extraction Algorithm & Compression Algorithm */  
/*      Bit exact C-language software implementation using ETSI basic operations library                 */
/*      3GPP TS 26.243 v 6.1.0 November 2004                                                                  */
/*      ANSI C code for the Fixed-Point Distributed Speech Recognition Extended Advanced Front-end       */
/*=======================================================================================================*/
/*---------------------------------------------------------------------------
 *
 * FILE NAME: error_handle.c
 * PURPOSE:   This file contains the main program for the error recovery 
 *	      of the DSR Extended Front-end features for the purpose
 *            of speech reconstruction.
 *
 * INPUT:     The program reads 1) HTK format file, 2) Pitch period file,
 *            3) Classification file and 4) Error status file that have been
 *            generated by the  ETSI ES 202 212 Decoding Algorithm C language
 *            software implementation. 
 *
 * OUTPUT:    The program outputs:
 *	      1) HTK format file containing recovered Mel-Cepstrum parameters
 *            2) Recovered pitch period file
 *            3) Recovered voicing class file
 *            4) Recovered log-energy file, though this info is contained
 *               int the output 1)
 *---------------------------------------------------------------------------*/




#include <stdio.h>
#include <stdlib.h>
#include <string.h>


int main(int iArgc, char *pcArgv[])
{

  int i;
  int iLen, iLen1, iLen2;
  int iMedianFlag = 0;

  FILE *fpInFtr, *fpInPit, *fpInCls, *fpInEs;
  FILE *fpOutPit, *fpOutCls, *fpOutLoge, *fpOutFtr;
 
  float pfFtrArray[16];
  float pfEnrgArray[3];
  float pfTemp[3];
  float fMedian, fTemp;
  float fEnrg, fPit;
  unsigned char cEs, cCls;

  if (iArgc != 9)
  {
    printf("Wrong number of arguments\n\n");
    printf("USAGE: error_handle ftr_in pit_in cls_in es_in ftr_out pit_out cls_out loge_out\n\n");
    printf("where\n");
    printf("     ftr_in:    input htk feature file\n");
    printf("     pit_in:    input pitch file\n");
    printf("     cls_in:    input class file\n");
    printf("     es_in:     input error status file\n");
    printf("     ftr_out:   output htk feature file\n");
    printf("     pit_out:   output pitch file\n");
    printf("     cls_out:   output class file\n");
    printf("     loge_out:  output log-E file\n\n");

    exit(1);
  }

  fpInFtr = fopen(pcArgv[1],"rb");
  if (NULL == fpInFtr)
  {
    printf("Failed to open input feature file\n");
    exit(1);
  }

  fpInPit = fopen(pcArgv[2],"rb");
  if (NULL == fpInPit)
  {
    printf("Failed to open input pitch file\n");
    exit(1);
  }

  fpInCls = fopen(pcArgv[3],"r");
  if (NULL == fpInCls)
  {
    printf("Failed to open input class file\n");
    exit(1);
  }

  fpInEs = fopen(pcArgv[4],"r");
  if (NULL == fpInEs)
  {
    printf("Failed to open input error status file\n");
    exit(1);
  }

  fpOutFtr = fopen(pcArgv[5],"w+b");
  if (NULL == fpOutFtr)
  {
    printf("Failed to open output feature file\n");
    exit(1);
  }
  
  fpOutPit = fopen(pcArgv[6],"wb");
  if (NULL == fpOutPit)
  {
    printf("Failed to open output pitch file\n");
    exit(1);
  }

  fpOutCls = fopen(pcArgv[7],"w");
  if (NULL == fpOutCls)
  {
    printf("Failed to open output class file\n");
    exit(1);
  }

  fpOutLoge = fopen(pcArgv[8],"wb");
  if (NULL == fpOutLoge)
  {
    printf("Failed to open output Log-E file\n");
    exit(1);
  }


  /*
   * First perform a median filtering on all the energy
   * values if errors are present.  Write out the filtered
   * energy values to a temporary file.
   */

  fread(pfFtrArray,sizeof(float),16,fpInFtr);
  fwrite(pfFtrArray,sizeof(float),16,fpOutFtr);

  fread(pfEnrgArray,sizeof(float),1,fpInFtr);
  fread(pfFtrArray,sizeof(float),13,fpInFtr);
  fread(pfEnrgArray+1,sizeof(float),1,fpInFtr);

  fscanf(fpInEs,"%c",&cEs);
  if (cEs != '0')
  {
    iMedianFlag = 1;
  }
  fscanf(fpInEs,"%c",&cEs);
  if (cEs != '0')
  {
    iMedianFlag = 1;
  }

  if (iMedianFlag)
  {
    fMedian = (pfEnrgArray[0]+pfEnrgArray[1])/2.0;
    fwrite(&fMedian,sizeof(float),1,fpOutFtr);
  }
  else
  {
    fwrite(pfEnrgArray,sizeof(float),1,fpOutFtr);
  }
  fwrite(pfFtrArray,sizeof(float),13,fpOutFtr);
 
  while(1)
  {

    fscanf(fpInEs,"%c",&cEs);
    if (cEs != '0')
    {
      iMedianFlag = 1;
    }

    fread(pfFtrArray,sizeof(float),13,fpInFtr);
    fread(pfEnrgArray+2,sizeof(float),1,fpInFtr);

    if (feof(fpInFtr))
    {
      break;
    }

    for (i = 0; i < 3; i++)
    {
      pfTemp[i] = pfEnrgArray[i];
    }
    if (pfTemp[1] < pfTemp[0])
    {
      fTemp = pfTemp[0];
      pfTemp[0] = pfTemp[1];
      pfTemp[1] = fTemp;
    }
    if (pfTemp[2] < pfTemp[0])
    {
      fTemp = pfTemp[0];
      pfTemp[0] = pfTemp[2];
      pfTemp[2] = fTemp;
    }
    if (pfTemp[2] < pfTemp[1])
    {
      fTemp = pfTemp[1];
      pfTemp[1] = pfTemp[2];
      pfTemp[2] = fTemp;
    }
    fMedian = pfTemp[1];

    if (iMedianFlag)
    {
      fwrite(&fMedian,sizeof(float),1,fpOutFtr);
    }
    else
    {
      fwrite(pfEnrgArray+1,sizeof(float),1,fpOutFtr);
    }
    fwrite(pfFtrArray,sizeof(float),13,fpOutFtr);

    pfEnrgArray[0] = pfEnrgArray[1];
    pfEnrgArray[1] = pfEnrgArray[2];

  }

  fMedian = (pfEnrgArray[0]+pfEnrgArray[1])/2.0;
  if (iMedianFlag)
  {
    fwrite(&fMedian,sizeof(float),1,fpOutFtr);
  }
  else
  {
    fwrite(pfEnrgArray+1,sizeof(float),1,fpOutFtr);
  }

  fclose(fpInFtr);
  fflush(fpOutFtr);

  rewind(fpOutFtr);
  rewind(fpInEs);


  /*
   * Depending on the length of the erasure, perform some
   * error concealment
   */

  if (0 != fseek(fpOutFtr,sizeof(float)*16,SEEK_SET))
  {
    printf("Error seeking output feature file\n");
    exit(1);
  }
 
  while(1)
  {

    fscanf(fpInEs,"%c",&cEs);

    if (feof(fpInEs))
    {
      break;
    }

    if (cEs == '0')
    {

      fread(&fEnrg,sizeof(float),1,fpOutFtr);
      fwrite(&fEnrg,sizeof(float),1,fpOutLoge);
      fseek(fpOutFtr,sizeof(float)*13,SEEK_CUR);

      fread(&fPit,sizeof(float),1,fpInPit);
      fwrite(&fPit,sizeof(float),1,fpOutPit);

      fscanf(fpInCls,"%c",&cCls);
      fprintf(fpOutCls,"%c",cCls);

    }

    else
    {

      i = 1;
      while (1)
      {

        fscanf(fpInEs,"%c",&cEs);
        if (feof(fpInEs))
        {
          break;
        }

        i++;

        if (cEs == '0')
        {
          break;
        }

      }

      if (cEs == '0')
      {
        iLen = i-1;
      }
      else
      {
        iLen = i;
      }

      if (iLen <= 4)
      {

        for (i = 0; i < iLen; i++)
        {
          fread(&fEnrg,sizeof(float),1,fpOutFtr);
          fwrite(&fEnrg,sizeof(float),1,fpOutLoge);
          fseek(fpOutFtr,sizeof(float)*13,SEEK_CUR);
 
          fread(&fPit,sizeof(float),1,fpInPit);
          fwrite(&fPit,sizeof(float),1,fpOutPit);

          fscanf(fpInCls,"%c",&cCls);
          fprintf(fpOutCls,"%c",cCls);
        }

        if (cEs == '0')
        {
          fread(&fEnrg,sizeof(float),1,fpOutFtr);
          fwrite(&fEnrg,sizeof(float),1,fpOutLoge);
          fseek(fpOutFtr,sizeof(float)*13,SEEK_CUR);
 
          fread(&fPit,sizeof(float),1,fpInPit);
          fwrite(&fPit,sizeof(float),1,fpOutPit);

          fscanf(fpInCls,"%c",&cCls);
          fprintf(fpOutCls,"%c",cCls);
        }

      }
      else
      {

        for (i = 0; i < 2; i++)
        {
          fread(&fEnrg,sizeof(float),1,fpOutFtr);
          fwrite(&fEnrg,sizeof(float),1,fpOutLoge);
          fseek(fpOutFtr,sizeof(float)*13,SEEK_CUR);
 
          fread(&fPit,sizeof(float),1,fpInPit);
          fwrite(&fPit,sizeof(float),1,fpOutPit);

          fscanf(fpInCls,"%c",&cCls);
          fprintf(fpOutCls,"%c",cCls);
        }

        if (iLen >= 24)
        {
          iLen1 = 12;
          iLen2 = iLen-12;
        }
        else
        {
          iLen1 = iLen/2;
          iLen2 = iLen/2;
        }

        for (i = 2; i < iLen1; i++)
        {
          fread(&fEnrg,sizeof(float),1,fpOutFtr);
          fEnrg = fEnrg - 2.0*(float)(i-1);
          if (fEnrg < 4.7)
          {
            fEnrg = 4.7;
          }
          fwrite(&fEnrg,sizeof(float),1,fpOutLoge);
          fseek(fpOutFtr,sizeof(float)*13,SEEK_CUR);
 
          fread(&fPit,sizeof(float),1,fpInPit);
          fwrite(&fPit,sizeof(float),1,fpOutPit);

          fscanf(fpInCls,"%c",&cCls);
          if (cCls == '3')
          {
            cCls = '2';
          }
          fprintf(fpOutCls,"%c",cCls);
        }

        for (i = iLen1; i < iLen/2; i++)
        {
          fread(&fEnrg,sizeof(float),1,fpOutFtr);
          fEnrg = 4.7;
          fwrite(&fEnrg,sizeof(float),1,fpOutLoge);
          fseek(fpOutFtr,sizeof(float)*13,SEEK_CUR);
 
          fread(&fPit,sizeof(float),1,fpInPit);
          fPit = 0.0;
          fwrite(&fPit,sizeof(float),1,fpOutPit);

          fscanf(fpInCls,"%c",&cCls);
          cCls = '1';
          fprintf(fpOutCls,"%c",cCls);
        }

        for (i = iLen/2; i < iLen2; i++)
        {
          fread(&fEnrg,sizeof(float),1,fpOutFtr);
          fEnrg = 4.7;
          fwrite(&fEnrg,sizeof(float),1,fpOutLoge);
          fseek(fpOutFtr,sizeof(float)*13,SEEK_CUR);
 
          fread(&fPit,sizeof(float),1,fpInPit);
          fPit = 0.0;
          fwrite(&fPit,sizeof(float),1,fpOutPit);

          fscanf(fpInCls,"%c",&cCls);
          cCls = '1';
          fprintf(fpOutCls,"%c",cCls);
        }

        for (i = iLen2; i < iLen-2; i++)
        {
          fread(&fEnrg,sizeof(float),1,fpOutFtr);
          fEnrg = fEnrg - 2.0*(float)(iLen-2-i);
          if (fEnrg < 4.7)
          {
            fEnrg = 4.7;
          }
          fwrite(&fEnrg,sizeof(float),1,fpOutLoge);
          fseek(fpOutFtr,sizeof(float)*13,SEEK_CUR);
 
          fread(&fPit,sizeof(float),1,fpInPit);
          fwrite(&fPit,sizeof(float),1,fpOutPit);

          fscanf(fpInCls,"%c",&cCls);
          if (cCls == '3')
          {
            cCls = '2';
          }
          fprintf(fpOutCls,"%c",cCls);
        }

        for (i = iLen-2; i < iLen ; i++)
        {
          fread(&fEnrg,sizeof(float),1,fpOutFtr);
          fwrite(&fEnrg,sizeof(float),1,fpOutLoge);
          fseek(fpOutFtr,sizeof(float)*13,SEEK_CUR);
 
          fread(&fPit,sizeof(float),1,fpInPit);
          fwrite(&fPit,sizeof(float),1,fpOutPit);

          fscanf(fpInCls,"%c",&cCls);
          fprintf(fpOutCls,"%c",cCls);
        }

        if (cEs == '0')
        {
          fread(&fEnrg,sizeof(float),1,fpOutFtr);
          fwrite(&fEnrg,sizeof(float),1,fpOutLoge);
          fseek(fpOutFtr,sizeof(float)*13,SEEK_CUR);
 
          fread(&fPit,sizeof(float),1,fpInPit);
          fwrite(&fPit,sizeof(float),1,fpOutPit);

          fscanf(fpInCls,"%c",&cCls);
          fprintf(fpOutCls,"%c",cCls);
        }

      }

    }

  }

  fclose(fpOutFtr);
  fclose(fpInPit);
  fclose(fpInCls);
  fclose(fpInEs);
  fclose(fpOutLoge);
  fclose(fpOutPit);
  fclose(fpOutCls);

  exit(0);
 
}



