/*=======================================================================================================*/
/*      Distributed Speech Recognition                                                                   */
/*      ETSI ES 202 050 Advanced Front-End Feature Extraction Algorithm & Compression Algorithm &        */
/*      ETSI ES 202 212 Extended Advanced Front-End Feature Extraction Algorithm & Compression Algorithm */  
/*      Bit exact C-language software implementation using ETSI basic operations library                 */
/*      3GPP TS 26.243 v 6.1.0 November 2004                                                                  */
/*      ANSI C code for the Fixed-Point Distributed Speech Recognition Extended Advanced Front-end       */
/*=======================================================================================================*/

#include <stdlib.h>
#include <stdio.h>
#include "ParmInterface_B.h"
#include "CompCepsExports_B.h"
#include "ff4nrFix16_B.h"
#include "MathFunc.h"
#include "typedef.h"
#include "basic_op.h"
#include "count.h"

#if defined(F_16KHZ)
#include "16kHzProcExports_B.h"
#endif


#define CC_PRE_EMPHASIS         0.90
#define NUM_CHANNELS_WI7        23
#define PRE_EMPHASIS16          ( (X_INT16) (CC_PRE_EMPHASIS *32768.0 + 0.5))
#define PRE_EMPHASIShalf16      ( (X_INT16) (CC_PRE_EMPHASIS *16384.0 + 0.5))
#define ENERGYFLOOR_FB          -10.0 

#define LOG_ENERGYFLOOR_FB -7387    /* ((X_INT32) (ENERGYFLOOR_FB/log(2.0) * 512 - 0.5))  */

#define CC_ENERGYFLOOR_logE -50.0


extern void FilterBank (X_INT32 *SigFFT,X_INT32 *intFilter) ;



/* local functions */
static void CepsCompute(X_INT16 *bufferIn,X_INT16 *cepOut,X_INT16 expWavProc
#if defined(F_16KHZ)
                        ,DataFor16kProc_B *pData16k
#endif
                        )  ;
static void PreEmphHamm (X_INT16 * bufferIn,X_INT16 *bufferOut) ;
static void CosInv (X_INT16 *iData,X_INT16 *cepOut) ;
static void CosInv16Khz (X_INT16 *iData16,X_INT16 *cepOut16) ;


/**************************** STATIC DATA  *********************************************/
#define CEP_FRAME_LENGTH        FRAME_LENGTH
#define CEP_FFT_LENGTH          256

#define CEP_NB_COEFS            (NUM_CEP_COEFF)
#define CEP_NB_CHANNELS         23


static const X_INT16 HalfHamming16[CEP_FRAME_LENGTH/2] = {  /* x 2**15 */
           2623  /* 0.54 - 0.46 * cos (PI * 2 * (0 + 0.5) / (200)*/
        ,  2638  /* 0.54 - 0.46 * cos (PI * 2 * (1 + 0.5) / (200)*/
        ,  2668  /* 0.54 - 0.46 * cos (PI * 2 * (2 + 0.5) / (200)*/
        ,  2712  /* 0.54 - 0.46 * cos (PI * 2 * (3 + 0.5) / (200)*/
        ,  2772  /* 0.54 - 0.46 * cos (PI * 2 * (4 + 0.5) / (200)*/
        ,  2846  /* 0.54 - 0.46 * cos (PI * 2 * (5 + 0.5) / (200)*/
        ,  2935  /* 0.54 - 0.46 * cos (PI * 2 * (6 + 0.5) / (200)*/
        ,  3038  /* 0.54 - 0.46 * cos (PI * 2 * (7 + 0.5) / (200)*/
        ,  3156  /* 0.54 - 0.46 * cos (PI * 2 * (8 + 0.5) / (200)*/
        ,  3288  /* 0.54 - 0.46 * cos (PI * 2 * (9 + 0.5) / (200)*/
        ,  3434  /* 0.54 - 0.46 * cos (PI * 2 * (10 + 0.5) / (200)*/
        ,  3595  /* 0.54 - 0.46 * cos (PI * 2 * (11 + 0.5) / (200)*/
        ,  3769  /* 0.54 - 0.46 * cos (PI * 2 * (12 + 0.5) / (200)*/
        ,  3957  /* 0.54 - 0.46 * cos (PI * 2 * (13 + 0.5) / (200)*/
        ,  4158  /* 0.54 - 0.46 * cos (PI * 2 * (14 + 0.5) / (200)*/
        ,  4373  /* 0.54 - 0.46 * cos (PI * 2 * (15 + 0.5) / (200)*/
        ,  4602  /* 0.54 - 0.46 * cos (PI * 2 * (16 + 0.5) / (200)*/
        ,  4843  /* 0.54 - 0.46 * cos (PI * 2 * (17 + 0.5) / (200)*/
        ,  5096  /* 0.54 - 0.46 * cos (PI * 2 * (18 + 0.5) / (200)*/
        ,  5363  /* 0.54 - 0.46 * cos (PI * 2 * (19 + 0.5) / (200)*/
        ,  5641  /* 0.54 - 0.46 * cos (PI * 2 * (20 + 0.5) / (200)*/
        ,  5931  /* 0.54 - 0.46 * cos (PI * 2 * (21 + 0.5) / (200)*/
        ,  6233  /* 0.54 - 0.46 * cos (PI * 2 * (22 + 0.5) / (200)*/
        ,  6546  /* 0.54 - 0.46 * cos (PI * 2 * (23 + 0.5) / (200)*/
        ,  6870  /* 0.54 - 0.46 * cos (PI * 2 * (24 + 0.5) / (200)*/
        ,  7205  /* 0.54 - 0.46 * cos (PI * 2 * (25 + 0.5) / (200)*/
        ,  7550  /* 0.54 - 0.46 * cos (PI * 2 * (26 + 0.5) / (200)*/
        ,  7905  /* 0.54 - 0.46 * cos (PI * 2 * (27 + 0.5) / (200)*/
        ,  8270  /* 0.54 - 0.46 * cos (PI * 2 * (28 + 0.5) / (200)*/
        ,  8644  /* 0.54 - 0.46 * cos (PI * 2 * (29 + 0.5) / (200)*/
        ,  9028  /* 0.54 - 0.46 * cos (PI * 2 * (30 + 0.5) / (200)*/
        ,  9419  /* 0.54 - 0.46 * cos (PI * 2 * (31 + 0.5) / (200)*/
        ,  9819  /* 0.54 - 0.46 * cos (PI * 2 * (32 + 0.5) / (200)*/
        , 10227  /* 0.54 - 0.46 * cos (PI * 2 * (33 + 0.5) / (200)*/
        , 10641  /* 0.54 - 0.46 * cos (PI * 2 * (34 + 0.5) / (200)*/
        , 11063  /* 0.54 - 0.46 * cos (PI * 2 * (35 + 0.5) / (200)*/
        , 11492  /* 0.54 - 0.46 * cos (PI * 2 * (36 + 0.5) / (200)*/
        , 11926  /* 0.54 - 0.46 * cos (PI * 2 * (37 + 0.5) / (200)*/
        , 12367  /* 0.54 - 0.46 * cos (PI * 2 * (38 + 0.5) / (200)*/
        , 12812  /* 0.54 - 0.46 * cos (PI * 2 * (39 + 0.5) / (200)*/
        , 13263  /* 0.54 - 0.46 * cos (PI * 2 * (40 + 0.5) / (200)*/
        , 13717  /* 0.54 - 0.46 * cos (PI * 2 * (41 + 0.5) / (200)*/
        , 14176  /* 0.54 - 0.46 * cos (PI * 2 * (42 + 0.5) / (200)*/
        , 14638  /* 0.54 - 0.46 * cos (PI * 2 * (43 + 0.5) / (200)*/
        , 15103  /* 0.54 - 0.46 * cos (PI * 2 * (44 + 0.5) / (200)*/
        , 15571  /* 0.54 - 0.46 * cos (PI * 2 * (45 + 0.5) / (200)*/
        , 16041  /* 0.54 - 0.46 * cos (PI * 2 * (46 + 0.5) / (200)*/
        , 16512  /* 0.54 - 0.46 * cos (PI * 2 * (47 + 0.5) / (200)*/
        , 16985  /* 0.54 - 0.46 * cos (PI * 2 * (48 + 0.5) / (200)*/
        , 17458  /* 0.54 - 0.46 * cos (PI * 2 * (49 + 0.5) / (200)*/
        , 17931  /* 0.54 - 0.46 * cos (PI * 2 * (50 + 0.5) / (200)*/
        , 18405  /* 0.54 - 0.46 * cos (PI * 2 * (51 + 0.5) / (200)*/
        , 18877  /* 0.54 - 0.46 * cos (PI * 2 * (52 + 0.5) / (200)*/
        , 19349  /* 0.54 - 0.46 * cos (PI * 2 * (53 + 0.5) / (200)*/
        , 19819  /* 0.54 - 0.46 * cos (PI * 2 * (54 + 0.5) / (200)*/
        , 20286  /* 0.54 - 0.46 * cos (PI * 2 * (55 + 0.5) / (200)*/
        , 20751  /* 0.54 - 0.46 * cos (PI * 2 * (56 + 0.5) / (200)*/
        , 21214  /* 0.54 - 0.46 * cos (PI * 2 * (57 + 0.5) / (200)*/
        , 21672  /* 0.54 - 0.46 * cos (PI * 2 * (58 + 0.5) / (200)*/
        , 22127  /* 0.54 - 0.46 * cos (PI * 2 * (59 + 0.5) / (200)*/
        , 22577  /* 0.54 - 0.46 * cos (PI * 2 * (60 + 0.5) / (200)*/
        , 23023  /* 0.54 - 0.46 * cos (PI * 2 * (61 + 0.5) / (200)*/
        , 23463  /* 0.54 - 0.46 * cos (PI * 2 * (62 + 0.5) / (200)*/
        , 23898  /* 0.54 - 0.46 * cos (PI * 2 * (63 + 0.5) / (200)*/
        , 24326  /* 0.54 - 0.46 * cos (PI * 2 * (64 + 0.5) / (200)*/
        , 24748  /* 0.54 - 0.46 * cos (PI * 2 * (65 + 0.5) / (200)*/
        , 25163  /* 0.54 - 0.46 * cos (PI * 2 * (66 + 0.5) / (200)*/
        , 25570  /* 0.54 - 0.46 * cos (PI * 2 * (67 + 0.5) / (200)*/
        , 25970  /* 0.54 - 0.46 * cos (PI * 2 * (68 + 0.5) / (200)*/
        , 26362  /* 0.54 - 0.46 * cos (PI * 2 * (69 + 0.5) / (200)*/
        , 26745  /* 0.54 - 0.46 * cos (PI * 2 * (70 + 0.5) / (200)*/
        , 27119  /* 0.54 - 0.46 * cos (PI * 2 * (71 + 0.5) / (200)*/
        , 27484  /* 0.54 - 0.46 * cos (PI * 2 * (72 + 0.5) / (200)*/
        , 27839  /* 0.54 - 0.46 * cos (PI * 2 * (73 + 0.5) / (200)*/
        , 28184  /* 0.54 - 0.46 * cos (PI * 2 * (74 + 0.5) / (200)*/
        , 28519  /* 0.54 - 0.46 * cos (PI * 2 * (75 + 0.5) / (200)*/
        , 28843  /* 0.54 - 0.46 * cos (PI * 2 * (76 + 0.5) / (200)*/
        , 29157  /* 0.54 - 0.46 * cos (PI * 2 * (77 + 0.5) / (200)*/
        , 29458  /* 0.54 - 0.46 * cos (PI * 2 * (78 + 0.5) / (200)*/
        , 29749  /* 0.54 - 0.46 * cos (PI * 2 * (79 + 0.5) / (200)*/
        , 30027  /* 0.54 - 0.46 * cos (PI * 2 * (80 + 0.5) / (200)*/
        , 30293  /* 0.54 - 0.46 * cos (PI * 2 * (81 + 0.5) / (200)*/
        , 30547  /* 0.54 - 0.46 * cos (PI * 2 * (82 + 0.5) / (200)*/
        , 30788  /* 0.54 - 0.46 * cos (PI * 2 * (83 + 0.5) / (200)*/
        , 31016  /* 0.54 - 0.46 * cos (PI * 2 * (84 + 0.5) / (200)*/
        , 31231  /* 0.54 - 0.46 * cos (PI * 2 * (85 + 0.5) / (200)*/
        , 31433  /* 0.54 - 0.46 * cos (PI * 2 * (86 + 0.5) / (200)*/
        , 31621  /* 0.54 - 0.46 * cos (PI * 2 * (87 + 0.5) / (200)*/
        , 31795  /* 0.54 - 0.46 * cos (PI * 2 * (88 + 0.5) / (200)*/
        , 31955  /* 0.54 - 0.46 * cos (PI * 2 * (89 + 0.5) / (200)*/
        , 32102  /* 0.54 - 0.46 * cos (PI * 2 * (90 + 0.5) / (200)*/
        , 32234  /* 0.54 - 0.46 * cos (PI * 2 * (91 + 0.5) / (200)*/
        , 32352  /* 0.54 - 0.46 * cos (PI * 2 * (92 + 0.5) / (200)*/
        , 32455  /* 0.54 - 0.46 * cos (PI * 2 * (93 + 0.5) / (200)*/
        , 32544  /* 0.54 - 0.46 * cos (PI * 2 * (94 + 0.5) / (200)*/
        , 32618  /* 0.54 - 0.46 * cos (PI * 2 * (95 + 0.5) / (200)*/
        , 32677  /* 0.54 - 0.46 * cos (PI * 2 * (96 + 0.5) / (200)*/
        , 32722  /* 0.54 - 0.46 * cos (PI * 2 * (97 + 0.5) / (200)*/
        , 32751  /* 0.54 - 0.46 * cos (PI * 2 * (98 + 0.5) / (200)*/
        , 32766  /* 0.54 - 0.46 * cos (PI * 2 * (99 + 0.5) / (200)*/
        } ;
        
        
/* inverse cosine transform coefficients  (x 2**15) */
static const X_INT16 CosMatrix16[(CEP_NB_COEFS -1) * ((CEP_NB_CHANNELS+1)/2)] = {
    22660,22238,21402,20167,18556,16600,14334,11801,9049,6128,3093,0                    /* C1  */
        ,22501,20833,17619,13098,7606,1550,-4621,-10449,-15503,-19406,-21871,-22714     /* C2  */
        ,22238,18556,11801,3093,-6128,-14334,-20167,-22660,-21402,-16600,-9049,0        /* C3  */
        ,21871,15503,4621,-7606,-17619,-22501,-20833,-13098,-1550,10449,19406,22714     /* C4  */
        ,21402,11801,-3093,-16600,-22660,-18556,-6128,9049,20167,22238,14334,0          /* C5  */
        ,20833,7606,-10449,-21871,-19406,-4621,13098,22501,17619,1550,-15503,-22714     /* C6  */
        ,20167,3093,-16600,-22238,-9049,11801,22660,14334,-6128,-21402,-18556,0         /* C7  */
        ,19406,-1550,-20833,-17619,4621,21871,15503,-7606,-22501,-13098,10449,22714     /* C8  */
        ,18556,-6128,-22660,-9049,16600,20167,-3093,-22238,-11801,14334,21402,0         /* C9  */
        ,17619,-10449,-21871,1550,22501,7606,-19406,-15503,13098,20833,-4621,-22714     /* C10 */
        ,16600,-14334,-18556,11801,20167,-9049,-21402,6128,22238,-3093,-22660,0         /* C11 */
        ,15503,-17619,-13098,19406,10449,-20833,-7606,21871,4621,-22501,-1550,22714     /* C12 */
} ;
#if defined(F_16KHZ)        
static const X_INT16 CosMatrix16_16khz[(CEP_NB_COEFS -1) * ((CEP_NB_CHANNELS+3+1)/2)] = {
         22672,22341,21685,20712,19437,17879,16061,14008,11750,9322,6757,4094,1371              /* C1  */
        ,22547,21237,18692,15062,10555,5436,0,-5436,-10555,-15062,-18692,-21237,-22547          /* C2  */
        ,22341,19437,14008,6757,-1371,-9322,-16061,-20712,-22672,-21685,-17879,-11750,-4094     /* C3  */
        ,22053,17001,8054,-2738,-12902,-20111,-22713,-20111,-12902,-2738,8054,17001,22053       /* C4  */
        ,21685,14008,1371,-11750,-20712,-22341,-16061,-4094,9322,19437,22672,17879,6757         /* C5  */
        ,21237,10555,-5436,-18692,-22547,-15062,0,15062,22547,18692,5436,-10555,-21237          /* C6  */
        ,20712,6757,-11750,-22341,-17879,-1371,16061,22672,14008,-4094,-19437,-21685,-9322      /* C7  */
        ,20111,2738,-17001,-22053,-8054,12902,22713,12902,-8054,-22053,-17001,2738,20111        /* C8  */
        ,19437,-1371,-20712,-17879,4094,21685,16061,-6757,-22341,-14008,9322,22672,11750        /* C9  */
        ,18692,-5436,-22547,-10555,15062,21237,0,-21237,-15062,10555,22547,5436,-18692          /* C10 */
        ,17879,-9322,-22341,-1371,21685,11750,-16061,-19437,6757,22672,4094,-20712,-14008       /* C11 */
        ,17001,-12902,-20111,8054,22053,-2738,-22713,-2738,22053,8054,-20111,-12902,17001       /* C12 */
        } ;
#endif      

/* mel bank coefficients */
static const X_INT16 pondMelFilter[] = {  /* x 2**11 */
      /* width first coeffs... */
            5,   2 ,5464,10920,16384,10920,5464 /* Channel 0 */
        ,   5,   4 ,5464,10920,16384,10920,5464 /* Channel 1 */
        ,   6,   6 ,5464,10920,16384,12288,8192,4096 /* Channel 2 */
        ,   6,   8 ,4096,8192,12288,16384,10920,5464 /* Channel 3 */
        ,   6,  11 ,5464,10920,16384,12288,8192,4096 /* Channel 4 */
        ,   7,  13 ,4096,8192,12288,16384,12288,8192,4096 /* Channel 5 */
        ,   7,  16 ,4096,8192,12288,16384,12288,8192,4096 /* Channel 6 */
        ,   8,  19 ,4096,8192,12288,16384,13104,9832,6552,3280 /* Channel 7 */
        ,   9,  22 ,3280,6552,9832,13104,16384,13104,9832,6552,3280 /* Channel 8 */
        ,   9,  26 ,3280,6552,9832,13104,16384,13104,9832,6552,3280 /* Channel 9 */
        ,   9,  30 ,3280,6552,9832,13104,16384,13104,9832,6552,3280 /* Channel 10 */
        ,  10,  34 ,3280,6552,9832,13104,16384,13656,10920,8192,5464,2728 /* Channel 11 */
        ,  11,  38 ,2728,5464,8192,10920,13656,16384,13656,10920,8192,5464,2728 /* Channel 12 */
        ,  12,  43 ,2728,5464,8192,10920,13656,16384,14040,11704,9360,7024,4680,2344 /* Channel 13 */
        ,  13,  48 ,2344,4680,7024,9360,11704,14040,16384,14040,11704,9360,7024,4680,2344 /* Channel 14 */
        ,  13,  54 ,2344,4680,7024,9360,11704,14040,16384,14040,11704,9360,7024,4680,2344 /* Channel 15 */
        ,  14,  60 ,2344,4680,7024,9360,11704,14040,16384,14336,12288,10240,8192,6144,4096,2048 /* Channel 16 */
        ,  16,  66 ,2048,4096,6144,8192,10240,12288,14336,16384,14560,12744,10920,9104,7280,5464,3640,1824 /* Channel 17*/
        ,  17,  73 ,1824,3640,5464,7280,9104,10920,12744,14560,16384,14560,12744,10920,9104,7280,5464,3640,1824 /* Channel 18 */
        ,  17,  81 ,1824,3640,5464,7280,9104,10920,12744,14560,16384,14560,12744,10920,9104,7280,5464,3640,1824 /* Channel 19 */
        ,  19,  89 ,1824,3640,5464,7280,9104,10920,12744,14560,16384,14896,13408,11912,10424,8936,7448,5960,4472,2976,1488 /* Channel 20 */
        ,  21,  97 ,1488,2976,4472,5960,7448,8936,10424,11912,13408,14896,16384,14896,13408,11912,10424,8936,7448,5960,4472,2976,1488 /* Channel 21 */
        ,  22, 107 ,1488,2976,4472,5960,7448,8936,10424,11912,13408,14896,16384,15016,13656,12288,10920,9560,8192,6824,5464,4096,2728,1368 /* Channel 22 */
        , 0 /* no more filter, length = 0 */
} ;
        
/****************************  local structures  *********************************************/
        
struct CompCepsStructX {
    int  FFTLength ;
#if defined(F_16KHZ)                /* 16 kHz processing */
    BOOLEAN Do16kHzProc;            /* do 16kHz processing ? */
    DataFor16kProc_B * pData16k;    /* data for 16Khz   */
#endif
} ;


/****************************  external functions  *********************************************/


/*----------------------------------------------------------------------------
 * FUNCTION NAME: DoCompCepsAlloc_B
 *
 * PURPOSE:  Allocation of a cepstral computing struct
 *
 * INPUT: N/A    
 *
 * OUTPUT: N/A
 *
 * RETURN VALUE: pointer to the new allocated struct
 *   
 *              
 *
 *---------------------------------------------------------------------------*/
extern CompCepsStructX * DoCompCepsAlloc_B (void) {
    CompCepsStructX * newCep = calloc (1,sizeof(CompCepsStructX)) ;
    if(newCep == NULL) return NULL;
    return newCep ;
}


/*----------------------------------------------------------------------------
 * FUNCTION NAME: DoCompCepsInit_B
 *
 * PURPOSE:  Initialisation of cepstral computing struct
 *
 * INPUT:
 *    - This : global FEParamsX_F context (for this->CCX)
 *
 * OUTPUT:
 *    - initialized this->CCX
 *
 * RETURN VALUE: none
 *   
 *              
 *
 *---------------------------------------------------------------------------*/


extern void DoCompCepsInit_B(
    FEParamsX_F *This       /* (i) Pointer to the FEParamsX_F containing the
                                   CompCepsStructX to initialize */
    ) {
    CompCepsStructX *CCX = This->CCX;
#if defined(F_16KHZ)
    CCX->Do16kHzProc     = This->Do16kHzProc;
    CCX->pData16k        = (DataFor16kProc_B *) This->pData16k;
#endif
    CCX->FFTLength = This->FFTLength ;
}



/*----------------------------------------------------------------------------
 * FUNCTION NAME: DoCompCeps_B
 *
 * PURPOSE: Cepstral computation for a frame 
 *              
 *
 * INPUT:
 *      - DataIn,   input signal frame
 *      - shiftIn,  Shift factor for DataIn[i]  DataIn[i] =Data[i] * 2**shiftIn
 *      - This  : global FEParamsX_F context (for this->CCX)                
 *    Warning : DataIn must contains the preceeding sample (index=-1) for preemphasis
 *
 * OUTPUT:
 *      - Coef16  :   output Cepstral ceofficient frame
 *    Warning : Size of Coef16 must be NUM_CEP_COEFF + 1 to receive c1 c2 ... cn c0 LogEnergy
 *
 * RETURN VALUE: TRUE
 *   
 *
 *---------------------------------------------------------------------------*/

extern BOOLEAN DoCompCeps_B (
    X_INT16 *DataIn,    /* (i) Current frame */
    X_INT16 shiftIn,    /* (i) Shift factor for DataIn[i]   */
    X_INT16 *Coef16,    /* (o) Output cesptral frame        */
    FEParamsX_F *This   /* (i) FEParamsX_F  struct          */
    ) 
{
    CompCepsStructX *CCX = This->CCX;
#if defined(F_16KHZ)
    DataFor16kProc_B *pData16k = CCX->pData16k;
#endif
    /* cepstral coefficients calculation */
    CepsCompute(DataIn,Coef16,shiftIn

#if defined(F_16KHZ)
        , (CCX->Do16kHzProc ? pData16k : NULL) 
#endif

        )  ;
    return TRUE;
}


/*----------------------------------------------------------------------------
 * FUNCTION NAME: DoCompCepsDelete_B
 *
 * PURPOSE:  frees a  cepstral computing struct
 *
 * INPUT:
 *    - CCX : cepstral computing struct
 *
 * OUTPUT: n/a
 *
 * RETURN VALUE: none
 *   
 *              
 *
 *---------------------------------------------------------------------------*/

extern void DoCompCepsDelete_B(
        CompCepsStructX *CCX    /* (i) Pointer to the structure */
) {
    if(CCX != NULL) {
        free(CCX);
    }
    return  ;
}


/***************************** local functions ******************************************/
/*----------------------------------------------------------------------------
 * FUNCTION NAME: CepsCompute
 *
 * PURPOSE: Cepstral computation for a frame 
 *              
 *
 * INPUT:
 *      - bufferIn16,   input signal frame
 *      - expWavProc,   Shift factor for bufferIn16[i]  bufferIn16[i] =bufferIn[i] * 2**expWavProc
 *      - pData16k ,    16Khz context struct (only for 16 Khz version)
 *
 *    WARNING : bufferIn16 must contains the preceeding sample (index=-1) for preemphasis
 *
 * OUTPUT:
 *      - cepOut,       output Cepstral coefficient frame
 *    WARNING : Size of cepOut must be NUM_CEP_COEFF + 1 to receive c1 c2 ... cn c0 LogEnergy
 *
 * RETURN VALUE: none
 *   
 *
 *---------------------------------------------------------------------------*/

static void CepsCompute(
        X_INT16 *bufferIn16,    /* (i)  input signal frame      */
        X_INT16 *cepOut,        /* (o)  output cepstral frame   */
        X_INT16 expWavProc      /* (i)  shift factor = log2(factor) of input signal */
#if defined(F_16KHZ)
       ,DataFor16kProc_B *pData16k  /* (i) pointer to 16Khz data        */
#endif
)  
{
    int i ;
    X_INT16 energy16  ;
    X_INT32 intBuffer32[CEP_FFT_LENGTH/2 +1] ;
	X_INT16 *intBuffer16 = (X_INT16 *)(intBuffer32) ; 
    X_INT32 intFilter32[CEP_NB_CHANNELS] ;
    X_INT16 * intFilter16 = (X_INT16 *)intFilter32 ;

#if defined(F_16KHZ)
    /*-------------------
    * 16 kHz processing
    *-------------------*/
    X_INT16   hp16kBandsSize;
    X_INT32   LogEnergy_B;
    X_INT16   eng_dec;
    X_INT32   hp16kBands_B[3];

    X_INT16   *codeWeights_B ;
    X_INT32   *CodeForBands16k_B;
    BOOLEAN   Do16kHzProc ;
    
    
    if (pData16k)   {
        hp16kBandsSize  = HP16k_MEL_USED;
        codeWeights_B     = Get16k_p_bufferCodeWeights_B (pData16k);  move16();
        CodeForBands16k_B = Get16k_p_bufferCodeForBands16k_B (pData16k);move16();
        Do16kHzProc = TRUE ;
    }   else    {
        hp16kBandsSize  = 0;
        codeWeights_B     = NULL;
        CodeForBands16k_B = NULL;
        Do16kHzProc = FALSE ;
    }
#endif

    {   /* log(ener) calculation */
        X_INT32 sum32 = L_mult(0,0) ;
        {
            for(i=0,sum32=0;i<CEP_FRAME_LENGTH;i+=8) {
                X_INT32 s0 , s1 , s2 ;
                s0 = L_mult(bufferIn16[i],bufferIn16[i]) ;
                s0 = L_mac(s0,bufferIn16[i+1],bufferIn16[i+1]);
                s1 = L_shr(s0,2) ;
                s0 = L_mult(bufferIn16[i+2],bufferIn16[i+2]);
                s0 = L_mac(s0,bufferIn16[i+3],bufferIn16[i+3]);
                s1 = L_add(s1,L_shr(s0,2)) ;
                s0 = L_mult(bufferIn16[i+4],bufferIn16[i+4]) ;
                s0 = L_mac(s0,bufferIn16[i+5],bufferIn16[i+5]);
                s2 = L_shr(s0,2) ;
                s0 = L_mult(bufferIn16[i+6],bufferIn16[i+6]);
                s0 = L_mac(s0,bufferIn16[i+7],bufferIn16[i+7]);
                s2 = L_add(s2,L_shr(s0,2)) ;
                sum32 = L_add(sum32,L_shr_r(L_add(s1,s2),6));
            }
        }
        energy16 = (X_INT16)  add(shr_r(Log_2(sum32),4), sub(7 * 64,shl(expWavProc,7)));
#if defined(F_16KHZ)
        if (Do16kHzProc) {
            LogEnergy_B = sum32; move32();
            eng_dec = sub(shl(expWavProc,1),7);
        }
#endif
    }
    /* preemphasis and hamming windowing */
    PreEmphHamm (bufferIn16,intBuffer16);
    /* Zero padding */
    for (i = CEP_FRAME_LENGTH; i < CEP_FFT_LENGTH; i++) {
        intBuffer16[i] = 0; move16();
    }
    /* inplace FFT */
    ff4NR16_B(8,intBuffer16);

    /* compute modules */
    intBuffer32[CEP_FFT_LENGTH / 2] = L_mult(intBuffer16[1],intBuffer16[1]); move32(); // first done because intBuffer32 and intBuffer16 overlap
    intBuffer32[0] = L_mult(intBuffer16[0],intBuffer16[0]); move32();
    for (i = 1; i < CEP_FFT_LENGTH / 2; i++)  { /* complex bins */
        intBuffer32[i] = L_mac(L_mult(intBuffer16[2*i],intBuffer16[2*i]),intBuffer16[2*i+1],intBuffer16[2*i+1]) ; move32();
    }

#if defined(F_16KHZ)
   /*-------------------
   * 16 kHz processing
   *-------------------*/
    if (Do16kHzProc)
    {               
    X_INT32 aux1 [3];
    X_INT16 BFD_dec[1];
    sub(shl(expWavProc,1),9);
        GetBandsForDecoding16k_B (intBuffer32, 2*expWavProc-9, aux1, BFD_dec, CEP_FFT_LENGTH/2+1);
        for (i=0 ; i<3 ; i++)
        {
            test();
            if (L_sub(aux1 [i],1) > 0)   {/* EnergyFloor_FB */
                aux1 [i] = L_sub(Log_2 ( aux1 [i]), shl((*BFD_dec),10)); move32();
            }else {
                aux1[i] = 1 ; move32();
            }
        }
        /* temporary */
        codeWeights_B [0] = (X_INT16)(0.1*(1<<15) + 0.5); move16();
        codeWeights_B [1] = (X_INT16)(0.2*(1<<15) + 0.5); move16() ;
        codeWeights_B [2] = (X_INT16)(0.7*(1<<15) + 0.5); move16() ;
        DecodeBands16k_B (hp16kBands_B, aux1, CodeForBands16k_B, codeWeights_B);
    }
#endif

/* mel scale filter bank */
    FilterBank (intBuffer32,intFilter32);
    /* low threshold for filters */
    {
        X_INT16 lowThres = add(LOG_ENERGYFLOOR_FB-10*512,shl(expWavProc,10)) ;
        for (i = 0; i < CEP_NB_CHANNELS; i++) {
            intFilter16[i] = shr_r(Log_2 ( intFilter32[i]),1) ; move16();
            test();
            if(sub(intFilter16[i],lowThres)  < 0 ) {
                intFilter16[i] = lowThres ; move16();
            }
        
        }
    }

#if defined(F_16KHZ)    
  /*-------------------
   * 16 kHz processing
   *-------------------*/
    if (Do16kHzProc)  
    {
        X_INT32 *hpBands_B ;
        X_INT32 IntBuffer[4];
        const X_INT16 hpBands_dec = Get16k_hpBands_dec_B (pData16k);
        const X_INT32 preemphs_correction_B = 948;    /* Q10 */
        hpBands_B = Get16k_p_hpBands_B (pData16k);
        /*-----------------------------------------------------------
        * SS HP bands -> including and rough preemphasis correction
        *-----------------------------------------------------------*/
        for (i=1 ; i<1+hp16kBandsSize ; i++){
            
            X_INT16 low = extract_l(hpBands_B[i-1]) ;
            X_INT16 high  = extract_h(hpBands_B[i-1]) ;
            const X_INT16 a16 = 31130;    //Q14
            high = sub(high,shr(low,15)) ;
            IntBuffer [i] = L_shl(L_mac(L_mult(a16,high),mult_r(a16,shr_r(low,1)),1),1) ; move32();
        }
        /*------------------------------------------------
        * coded HP bands -> rough preemphasis correction
        *------------------------------------------------*/
        for (i=0 ; i<hp16kBandsSize ; i++) {
            hp16kBands_B [i] = L_add(hp16kBands_B[i], preemphs_correction_B); move32();
        }
        /*-------------------------------
         * Natural logarithm computation
         *-------------------------------*/
        for (i=1 ; i<1+hp16kBandsSize ; i++) {
            test();
            if (L_sub(IntBuffer[i],1) < 0 ) {
                /* ROUND32() could have used intead */
                IntBuffer[i] = (ENERGYFLOOR_FB*1477.32 + 0.5); move32();
            } else {
                IntBuffer[i] = L_msu(L_deposit_l(Log_2(IntBuffer[i])),hpBands_dec,512); move32();
            }
        }
        
        /* add the last 8khz filter for Merge   */    
        IntBuffer[0] = L_msu(L_mult(intFilter16[CEP_NB_CHANNELS-1],1),1024,sub(expWavProc,5)) ; move32();
        /*--------------------------------------------
        * mixing coded and SS HP bands in log domain
        *--------------------------------------------*/
        MergeSSandCoded_B (IntBuffer, hp16kBands_B, 1, &hp16kBandsSize, pData16k);

        LogEnergy_B = CorrectEnergy_B (LogEnergy_B, &eng_dec, hp16kBandsSize, IntBuffer,1, CC_PRE_EMPHASIS); move32();
        test();
        if (L_sub(LogEnergy_B,1) < 0) {
            /* CC_ENERGYFLOOR_logE; */ /*LogEnergy_B = (-50*1477.32); */
            cepOut[ CEP_NB_COEFS ] = -50 * 64 ; move16()  ;
        } else {
            LogEnergy_B = L_sub(Log_2 (LogEnergy_B), shl(eng_dec,10)); /* insert energy with correct shift */
            /* remove log(2) */
			{
				X_INT16 low = extract_l(LogEnergy_B) ;
				X_INT16 high  = extract_h(LogEnergy_B) ;
				high = sub(high,shr(low,15)) ;
				/* 0x58c =  a16 = ROUND16(64/1477.32 * 32768)
				   0x90ae = (64/1477.32 * 32768 - a16) * 32768 (16b)
				   0xc857 = 0x90ae >> 1 
				*/
				cepOut[ CEP_NB_COEFS ] = add( extract_h((L_mac(L_mac(L_mac(0x8000,0x58c,low),0x90ae,high),1,mult(0xC857,low)))), extract_l(L_mult(0x58c,high))); move16();
			}
        }
        
        for(i=0;i<1+hp16kBandsSize ; i++) {
            intFilter16[i+CEP_NB_CHANNELS-1] = round(L_shl(L_mac(IntBuffer[i],1024,sub(expWavProc,5)),15)) ; move16();
        }
        CosInv16Khz (intFilter16,cepOut);
        /* remove expWavProc for C0 */
        /* 4141 =  2 * (23+3) * 32 * log(2) * 32 - 32678 */     
        cepOut[CEP_NB_COEFS-1] = sub(sub(cepOut[CEP_NB_COEFS-1],shl(expWavProc,10)),mult_r(4141,shl(expWavProc,10))) ; move16();
    } 
    else 
#endif  //defined(F_16KHZ)
    {
        CosInv (intFilter16,cepOut);
        /* remove expWavProc for C0 */
        cepOut[CEP_NB_COEFS-1] = sub(cepOut[CEP_NB_COEFS-1],mult_r(32650,shl(expWavProc,10))) ; move16();
        /* remove log(2) for energy */
        /* 22713 = log(2) * 32768 */
        cepOut[ CEP_NB_COEFS ] = mult_r( 22713,energy16) ; move16() ;
    }
    return ;
}


/*----------------------------------------------------------------------------
 * FUNCTION NAME: PreEmphHamm
 *
 * PURPOSE: preemphasis and Hamming windowing
 *              
 *
 * INPUT:
 *      - bufferIn16,   input signal frame
 *      - pData16k : 16Khz context struct (only for 16 Khz version)
 *    WARNING : bufferIn16 must contains the preceeding sample (index=-1) for preemphasis
 *
 * OUTPUT:
 *      - bufferOut16  :   output preemphasized and windowed signal 
 *
 * RETURN VALUE: none
 *   
 *
 *---------------------------------------------------------------------------*/

static void PreEmphHamm (
        X_INT16 * bufferIn16,   /* (i)  input signal */
        X_INT16 *bufferOut16    /* (o)  output premphasized and windowed signal */
    ) {
    int i;
    for (i = 0; i < CEP_FRAME_LENGTH/ 2; i++) {
        X_INT16 val = round(L_mac(L_mult(0x4000,bufferIn16[i]),-PRE_EMPHASIShalf16,bufferIn16[i-1]));
        bufferOut16[i] = mult_r(HalfHamming16[i],val);  move16();  
    }
    for (i = CEP_FRAME_LENGTH / 2; i < CEP_FRAME_LENGTH; i++) {
        X_INT16 val = round(L_mac(L_mult(0x4000,bufferIn16[i]),-PRE_EMPHASIShalf16,bufferIn16[i-1]));
        bufferOut16[i] = mult_r(HalfHamming16[CEP_FRAME_LENGTH - 1 - i],val);    move16();
    }
}

/*----------------------------------------------------------------------------
 * FUNCTION NAME: FilterBank
 *
 * PURPOSE: Mel filtering
 *              
 *
 * INPUT:
 *      - SigFFT32   FFT result input
 *
 * OUTPUT:
 *      - intFilter32  :  energy filter bank  
 *
 * RETURN VALUE: none
 *   
 *
 *---------------------------------------------------------------------------*/
extern void FilterBank (
        X_INT32 *SigFFT32,          /* (i) FFT result */
        X_INT32 *intFilter32        /* (o) filter  bank Energy  */ 
        ) {
    const X_INT16 *ptPond16  = pondMelFilter ;
    int nf = 0 ;
    int length ;
    test(); move16(); move16();
    while((length = *ptPond16++) > 0) { /* width */
        X_INT16 first;      /* first frequency */
        X_INT32 sum;  
        int j ; 

        first = *ptPond16++ ; move16(); 
        sum = 0 ;  move32();

        for(j=0;j<length;j++) {
            X_INT32 sf = SigFFT32[first+j] ;
            X_INT16 low = shr(extract_l(sf),1) ;
            X_INT32 high = extract_h(L_msu(sf,1,low));
            sum=L_mac(L_mac(sum,*ptPond16,high),1,mult_r(*ptPond16,low));
            ptPond16++ ; 
        }
        intFilter32[nf++] = sum    ; move32();
        test(); move16();
    }
    return;
}

/*----------------------------------------------------------------------------
 * FUNCTION NAME: CosInv
 *
 * PURPOSE: DCT for cepstrum calculation at 8Khz
 *              
 *
 * INPUT:
 *      - iData16    log energy filter bank input
 *
 * OUTPUT:
 *      - cepOut16  :  cepstral coefficients output (c1,c2,...ck,c0)  
 *
 * RETURN VALUE: none
 *   
 *
 *---------------------------------------------------------------------------*/
static void CosInv (
    X_INT16 *iData16,           /* (i) filter log energy */
    X_INT16 *cepOut16) {        /* (o) cepstral coefficients */
    int i;
    /* compute symetrical sum and diff  (/4) */
    for(i=0;i<CEP_NB_CHANNELS/2;i++) {
        X_INT16 sum = mac_r(L_mult(0x2000,iData16[i]),0x2000, iData16[CEP_NB_CHANNELS - i - 1]);
        iData16[CEP_NB_CHANNELS - i - 1] = msu_r(L_mult(0x2000,iData16[i]),0x2000,iData16[CEP_NB_CHANNELS - i - 1]) ; move16();
        iData16[i] = sum ; move16();
    }
    /* median value, applies only division by 4 */
    iData16[i] = shr_r(iData16[i],2); move16();
    
    { /* c1 ... ck */ /* CEP_NB_COEFS is even */
        const X_INT16 * ptCos = CosMatrix16 ;
        X_INT16 * ptFilt = iData16 + CEP_NB_CHANNELS - 1 ;
        move16();move16();
        for (i = 0; i < CEP_NB_COEFS -1;)
        {
            X_INT32 sum = L_mult(0,0) ;
            int j ;
            for (j = 0; j < (CEP_NB_CHANNELS+1)/2 ; j++) {
                X_INT16 pt1 = *ptCos++ ;
                sum  = L_mac(sum,ptFilt[-j],pt1) ; 
            }
            ptFilt -= CEP_NB_CHANNELS -1 ; move16();
            cepOut16[i++] = round(L_shl(sum,1)); move16();
            sum = L_mult(0,0) ;
            for (j = 0; j < (CEP_NB_CHANNELS+1)/2 ; j++) {
                X_INT16 pt1 = *ptCos++ ;
                sum  = L_mac(sum,ptFilt[j],pt1) ; 
            }
            ptFilt += CEP_NB_CHANNELS -1 ; move16();
            cepOut16[i++] = round(L_shl(sum,1)); move16();
        }
    }

/* Computing c0, as the last element of output vector */
#define LOG2_2xx15  22713 /* 32768 * log(2)) */
    {
        int j ;
        X_INT32 sum = (( CEP_NB_CHANNELS * 128 *2 * 10) << 13) + 0x8000  ; move32();
        for (j = 0; j < (CEP_NB_CHANNELS/2 ); j++) {
            sum = L_mac(sum,0x2000,iData16[j]) ;
        }
        sum = L_mac(sum,0x2000,iData16[j]) ;   /* CEP_NB_CHANNELS is odd */  
        
        cepOut16[CEP_NB_COEFS- 1] = mult_r(LOG2_2xx15,extract_h(sum))   ; move16();
    }
    
    return;
}

#if defined(F_16KHZ)
#define CEP_NB_CHANNELS_16Khz (CEP_NB_CHANNELS+3)


/*----------------------------------------------------------------------------
 * FUNCTION NAME: CosInv16Khz
 *
 * PURPOSE: DCT for cepstrum calculation at 16Khz
 *              
 *
 * INPUT:
 *      - iData16    log energy filter bank input
 *
 * OUTPUT:
 *      - cepOut16  :  cepstral coefficients output (c1,c2,...ck,c0)  
 *
 * RETURN VALUE: none
 *   
 *
 *---------------------------------------------------------------------------*/

void CosInv16Khz (
        X_INT16 *iData16    /* (i) filter log energy */
        ,X_INT16 *cepOut16  /* (o) cepstral coefficients */
) {
    int i;
    /* compute symetrical sum and diff  (/4) */
    for(i=0;i<CEP_NB_CHANNELS_16Khz/2;i++) {
        X_INT16 sum = mac_r(L_mult(0x2000,iData16[i]),0x2000, iData16[CEP_NB_CHANNELS_16Khz - i - 1]);
        iData16[CEP_NB_CHANNELS_16Khz - i - 1] = msu_r(L_mult(0x2000,iData16[i]),0x2000,iData16[CEP_NB_CHANNELS_16Khz - i - 1]) ; move16();
        iData16[i] = sum ; move16();
    }

    
    { /* c1 ... ck */ /* CEP_NB_COEFS is even */
        const X_INT16 * ptCos = CosMatrix16_16khz ;
        X_INT16 * ptFilt = iData16 + CEP_NB_CHANNELS_16Khz - 1 ;
        move16();move16();
        for (i = 0; i < CEP_NB_COEFS -1;)
        {
            X_INT32 sum = L_mult(0,0) ;
            int j ;
            for (j = 0; j < (CEP_NB_CHANNELS_16Khz)/2 ; j++) {
                X_INT16 pt1 = *ptCos++ ;
                sum  = L_mac(sum,ptFilt[-j],pt1) ;
            }
            ptFilt -= CEP_NB_CHANNELS_16Khz -1 ; move16();
            cepOut16[i++] = round(L_shl(sum,1)); move16();
            sum = L_mult(0,0) ;
            for (j = 0; j < (CEP_NB_CHANNELS_16Khz)/2 ; j++) {
                X_INT16 pt1 = *ptCos++ ;
                sum  = L_mac(sum,ptFilt[j],pt1) ;
            }
            ptFilt += CEP_NB_CHANNELS_16Khz -1 ; move16();
            cepOut16[i++] = round(L_shl(sum,1)); move16();
        }
    }

#define LOG2_2xx15  22713 /* 32768 * log(2)) */
    {
        int j ;
        X_INT32 sum = (( CEP_NB_CHANNELS_16Khz * 128 *2 * 10) << 13) + 0x8000  ; move32();
        for (j = 0; j < (CEP_NB_CHANNELS_16Khz/2 ); j++) {
            sum = L_mac(sum,0x2000,iData16[j]) ;
        }
        cepOut16[CEP_NB_COEFS- 1] = mult_r(LOG2_2xx15,extract_h(sum))   ; move16();
    }
    
    return;
}
#endif      

