/*=======================================================================================================*/
/*      Distributed Speech Recognition                                                                   */
/*      ETSI ES 202 050 Advanced Front-End Feature Extraction Algorithm & Compression Algorithm &        */
/*      ETSI ES 202 212 Extended Advanced Front-End Feature Extraction Algorithm & Compression Algorithm */  
/*      Bit exact C-language software implementation using ETSI basic operations library                 */
/*      3GPP TS 26.243 v 6.1.0 November 2004                                                                  */
/*      ANSI C code for the Fixed-Point Distributed Speech Recognition Extended Advanced Front-end       */
/*=======================================================================================================*/
/*
**********************************************************************
*                                                                    *
*                           Include Files                            *
*                                                                    *
**********************************************************************
*/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#include "classifyFrame_B.h"
#include "typedef.h"
#include "basic_op.h"
#include "count.h"


/*
**********************************************************************
*                                                                    *
*                               Defines                              *
*                                                                    *
**********************************************************************
*/

/*
 * Upper-Band energy fraction
 */

#define			UB_ENRG_FRAC		59


/*
 * Zero-crossing measure threshold
 */

#define			ZCM_THLD		87


/*
***********************************************************************
*
*   FUNCTION NAME: classify_frame
*
*   PURPOSE: To classify a frame into one of four classes
*
*        non-speech (0),
*        unvoiced speech (1),
*        mixed-voiced speech (2),
*        voiced speech (3)
*
*   INPUTS:
*
*      iVad - VAD value
*
*      iHangOverFlag - Hangover flag
*
*      lwPitchPeriod - Pitch period
*
*      nsFrameEnergy - Frame energy
*
*      iZcm          - Zero crossing measure
*
*      pswUBSpeech[] - Upper-Band speech
*
*   OUTPUTS:
*
*      None
*
*   RETURN VALUE:
*
*      The class of the frame is returned
*
**********************************************************************
*/


X_INT16 classify_frame(X_INT16 iVad,
                       X_INT16 iHangOverFlag,
                       Longword lwPitchPeriod,
                       NormSw nsFrameEnergy,
                       X_INT16 iZcm,
                       Shortword *pswUBSpeech)
{


  /*
  ********************************************************************
  *                                                                  *
  *                        Automatic variables                       *
  *                                                                  *
  ********************************************************************
  */

  X_INT16 i;
  X_INT16 iClass;
  X_INT16 iShift;
  X_INT16 iShift1;

  Shortword swMax;
  Shortword swTemp;

  Longword lwSumSq;
  Longword lwUBEnergy;


  /*
  ********************************************************************
  *                                                                  *
  *                         Executable Code                          *
  *                                                                  *
  ********************************************************************
  */


  /*
   * Set the default value
   */

  iClass = 3;						move16();


  /*
   * Estimate the class
   */

  test();
  if (iVad <= 0)
  {
    iClass = 0;						move16();
  }
  else 
  {
      test(); 
      if (extract_h(lwPitchPeriod) <= 0)
      {
          iClass = 1;						move16();
      }
      else
      {
          
          /*
           * Compute the upper-band energy and adjust it to the
           * same normalization level as the input frame energy
           */
          
          swMax = abs_s(pswUBSpeech[0]);
          for (i = 1; i < FRAME_LEN; i++)
          {
              swMax = swMax | abs_s(pswUBSpeech[i]);		logic16();
          }
          iShift = sub(norm_s(swMax),4);
          
          swTemp = shl(pswUBSpeech[0],iShift);
          lwSumSq = L_mult(swTemp,swTemp);
          for (i = 1; i < FRAME_LEN; i++)
          {
              swTemp = shl(pswUBSpeech[i],iShift);
              lwSumSq = L_mac(lwSumSq,swTemp,swTemp);
          }
          
          iShift1 = sub(add(shl(iShift,1),8),nsFrameEnergy.iShift);
          lwUBEnergy = L_shr(lwSumSq,iShift1);
          
          /* Decide */
          test(); 
          if ((L_msu(lwUBEnergy,nsFrameEnergy.swMantissa,UB_ENRG_FRAC) <= 0))
          {
              iClass = 2; move16();
          }
          else
          { 
              test(); 
              if(sub(iZcm, ZCM_THLD) >= 0)
              {
                  iClass = 2; move16();
              }
              else
              {
                  test(); 
                  if (iHangOverFlag != FALSE)
                  {
                      iClass = 2;					move16();
                  }
              }
          }
      }
  }
  return(iClass);

}



/*
***********************************************************************
*
*   FUNCTION NAME: get_zcm
*
*   PURPOSE: To obtain the zero crossing measure
*
*   INPUTS:
*
*      pswInpSpeech[] - Input speech
*
*   OUTPUTS:
*
*      None
*
*   RETURN VALUE:
*
*      The zero crossing measure is returned
*
**********************************************************************
*/

X_INT16 get_zcm(Shortword *pswInpSpeech)
{

  X_INT16 i;
  X_INT16 iSignPast;
  X_INT16 iSignPres;
  X_INT16 iZcm;


  /*
   * Compute the zero-crossing measure
   */

  iZcm = 0;						move16();
  test();
  iSignPast = (pswInpSpeech[0] >= 0)? +1 : -1;		move16();

  for (i = 1; i < FRAME_LEN; i++)
  {
    test();
    iSignPres = (pswInpSpeech[i] >= 0)? +1 : -1;	move16();
    test();
    if (sub(iSignPres,iSignPast) == 0)
    {
      ;
    }
    else
    {
      iZcm = add(iZcm,1);
    }
    iSignPast = iSignPres;				move16();
  }

  return(iZcm);

}

