/*=======================================================================================================*/
/*      Distributed Speech Recognition                                                                   */
/*      ETSI ES 202 050 Advanced Front-End Feature Extraction Algorithm & Compression Algorithm &        */
/*      ETSI ES 202 212 Extended Advanced Front-End Feature Extraction Algorithm & Compression Algorithm */  
/*      Bit exact C-language software implementation using ETSI basic operations library                 */
/*      3GPP TS 26.243 v 6.1.0 November 2004                                                                  */
/*      ANSI C code for the Fixed-Point Distributed Speech Recognition Extended Advanced Front-end       */
/*=======================================================================================================*/


/*
**********************************************************************
*                                                                    *
*                           Include Files                            *
*                                                                    *
**********************************************************************
*/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#include "computePCorr_B.h"
#include "fix_mathlib.h"
#include "typedef.h"
#include "basic_op.h"
#include "count.h"


/*
**********************************************************************
*                                                                    *
*                              Defines                               *
*                                                                    *
**********************************************************************
*/

#define		SQRT_ONE_HALF		0x5A82	/* 0.707 */
#define		FRAME_LEN_DS		50	/* 200/4 */
#define		FRAME_LEN_DS_BY_2	25
#define		HISTORY_LEN_DS		25	/* 100/4 */
#define		WINDOW_LENGTH		18
#define		INV_WINDOW_LENGTH	1820	/* 1/18 = 0.05556 */

Longword Mpy_lw_lw(Longword lwIn1, Longword lwIn2);
Longword Mpy_lw_sw(Longword lwIn1, Shortword swIn2);

typedef struct {
    Longword lwX1_X1;
    Longword lwZ1_Z1;
    Longword lwZ2_Z2;
    Longword lwX1_Z1;
    Longword lwX1_Z2;
    Longword lwZ1_Z2;

    Shortword swX1_Sum;
    Shortword swZ1_Sum;
    Shortword swZ2_Sum;
} PCORR_STATE_be;


/*
***********************************************************************
*
*   FUNCTION NAME: interpolate_be
*
*   PURPOSE: To interpolate correlation values
*
*   INPUTS:
*
*      *s_be - structure containing the correlations and sums
*
*      swAlpha - Alpha value
*
*      swBeta - Beta value
*
*   OUTPUTS:
*
*      None
*
*   RETURN VALUE:
*
*      The interpolated value is returned
*
**********************************************************************
*/

static Shortword interpolate_be(
                PCORR_STATE_be *s_be,
                Shortword swAlpha,
                Shortword swBeta
                )
  {
  Shortword swPCorr;
  Shortword swNumer, swDenom;
  Longword lwTmp, lwTmp2;
  X_INT16 iNumerShift;
  X_INT16 iDenomShift;
  X_INT16 iPCorrShift;

  lwTmp = Mpy_lw_sw(s_be->lwX1_Z1, swBeta);
  lwTmp2 = Mpy_lw_sw(s_be->lwX1_Z2, swAlpha);
  lwTmp = L_add(lwTmp, lwTmp2);
  iNumerShift = norm_l(lwTmp);
  iNumerShift = sub(iNumerShift, 1);
  lwTmp = L_shl(lwTmp, iNumerShift);
  swNumer = round(lwTmp);

  lwTmp = Mpy_lw_sw(s_be->lwZ1_Z1, swBeta);
  lwTmp = Mpy_lw_sw(lwTmp, swBeta);
  lwTmp2 = Mpy_lw_sw(s_be->lwZ1_Z2, swAlpha);
  lwTmp2 = Mpy_lw_sw(lwTmp2, swBeta);
  lwTmp2 = L_shl(lwTmp2, 1);
  lwTmp = L_add(lwTmp, lwTmp2);
  lwTmp2 = Mpy_lw_sw(s_be->lwZ2_Z2, swAlpha);
  lwTmp2 = Mpy_lw_sw(lwTmp2, swAlpha);
  lwTmp = L_add(lwTmp, lwTmp2);

  /* lwTmp in Q-29 */
  lwTmp = Mpy_lw_lw(lwTmp, s_be->lwX1_X1);
  iDenomShift = norm_l(lwTmp);
  iDenomShift = shr(iDenomShift, 1);
  iDenomShift = shl(iDenomShift, 1);
  lwTmp = L_shl(lwTmp, iDenomShift);
  swDenom = sqrt_l_fix(lwTmp);
  iDenomShift = shr(iDenomShift, 1);

  test();
  if (swDenom > 0)
  {
      test();
      if (swNumer > 0)
      {
          test();
          if (swNumer < swDenom)
          {
              swPCorr = div_s(swNumer, swDenom);
          }
          else
          {
              swPCorr = 0x7FFF;  move16();
          }
          
          iPCorrShift = sub(iNumerShift, iDenomShift);
          swPCorr = shr_r(swPCorr, iPCorrShift);
      }
      else
      {
          swPCorr = 0;			move16();
      }
      
  }
  return swPCorr;
  }

 
/*
***********************************************************************
*
*   FUNCTION NAME:
*
*      accumulate_be
*
*   PURPOSE:
*
*      To update the correlations and sums
*
*   INPUTS:
*
*      *pswSignal - Processed, downsampled speech input
*
*      iX1Index - Index into the speech buffer
*
*      iPitchPeriod - Pitch period
*
*      iWinLength - Window length
*
*   INPUT/OUTPUT:
*
*      *s - structure containing correlation and sums - this is updated
*
*   OUTPUTS:
*
*      None
*
*   RETURN VALUE:
*
*      None
*
**********************************************************************
*/

static void accumulate_be(PCORR_STATE_be *s,            
                       Shortword *pswSignal,
                       X_INT16 iX1Index,
                       X_INT16 iPitchPeriod,
                       X_INT16 iWinLength)
  {
  X_INT16 j;
  X_INT16 iZ1Index;
  X_INT16 iZ2Index;

  iZ1Index = sub(iX1Index,iPitchPeriod);
  iZ2Index = add(iZ1Index,1);

  /*
     the following values are cached in local variables before
     being used (without penalty) in the loop. The caching
     procedure requires some move16() and move32():
      s->swX1_Sum, s->swZ1_Sum, s->lwX1_X1, s->lwZ1_Z1, 
      s->lwX1_Z1, s->lwX1_Z2, s->lwZ1_Z1
  */
  move16(); move16(); 
  move32(); move32(); move32(); move32(); move32();

  for (j = 0; j < iWinLength; j++)
    {
    s->swX1_Sum = add(s->swX1_Sum,pswSignal[iX1Index+j]);
    s->swZ1_Sum = add(s->swZ1_Sum,pswSignal[iZ1Index+j]);

    s->lwX1_X1 = L_mac(s->lwX1_X1,pswSignal[iX1Index+j],pswSignal[iX1Index+j]);
    s->lwZ1_Z1 = L_mac(s->lwZ1_Z1,pswSignal[iZ1Index+j],pswSignal[iZ1Index+j]);

    s->lwX1_Z1 = L_mac(s->lwX1_Z1,pswSignal[iX1Index+j],pswSignal[iZ1Index+j]);
    s->lwX1_Z2 = L_mac(s->lwX1_Z2,pswSignal[iX1Index+j],pswSignal[iZ2Index+j]);
    s->lwZ1_Z2 = L_mac(s->lwZ1_Z2,pswSignal[iZ1Index+j],pswSignal[iZ2Index+j]);
    }

  s->swZ2_Sum = add(s->swZ1_Sum,sub(pswSignal[iZ1Index+j],
	pswSignal[iZ1Index]));                               move16();
  s->lwZ2_Z2 = L_add(s->lwZ1_Z1,L_msu
      (L_mult(pswSignal[iZ1Index+j],pswSignal[iZ1Index+j]),
       pswSignal[iZ1Index],pswSignal[iZ1Index]));            move32();

  /* inverse caching procedure */
  move16(); move16(); 
  move32(); move32(); move32(); move32(); move32();

  }


/*
***********************************************************************
*
*   FUNCTION NAME:
*
*      find_most_energetic_window2_be
*
*   PURPOSE:
*
*      To find the most energetic window simultaneously for two signals
*
*   INPUTS:
*
*      *pswSig1 - Signal number 1
*
*      *pswSig2 - Signal number 2
*
*      iLength - Pitch period length
*
*      iWinLength - Window length of the signals
*
*   OUTPUTS:
*
*      None
*
*   RETURN VALUE:
*
*      The most energetic position in the window is returned
*
**********************************************************************
*/
/* ------------------------------------------------------------------   
  Find the most energetic position of the window of given
  length (iWinLength) simultaneously for two signals (pfSig1 and pfSig2).
  Assume that each signal is periodic (with period length iLength), and
  is given by one period. 
 ------------------------------------------------------------------- */
static int find_most_energetic_window2_be(
                      Shortword  *pswSig1,
                      Shortword  *pswSig2,
                      int    iLength,
                      int    iWinLength
                      )
  {
  Longword lwSum, lwMax;
  X_INT16 i1, i2, iLoc, iLim;

  lwSum = 0; 			move32();
  iLim = iWinLength; 		move16();


  /* --------------------------------------------------------
     Initial window location

     i1             i2
     |                                           |
     |--------------|----------------------------|   
     |                                           | 

     -------------------------------------------------------- */
  for (i2=0; i2 < iLim; i2++)
    {
    lwSum = L_mac(lwSum, pswSig1[i2], pswSig1[i2]);
    lwSum = L_mac(lwSum, pswSig2[i2], pswSig2[i2]);
    }    
  lwMax = lwSum;		move32();
  iLoc = 0;			move16();


  /*
     Move the window up to the leftmost position
     i1             i2
     |         --------->                        |
     |----------------------------|--------------|   
     |                                           | 

   */
  i1=0;				move16();
  for ( ; i2 < iLength; i2++)
    {
    lwSum = L_mac(lwSum, pswSig1[i2], pswSig1[i2]);
    lwSum = L_msu(lwSum, pswSig1[i1], pswSig1[i1]);
    lwSum = L_mac(lwSum, pswSig2[i2], pswSig2[i2]);
    lwSum = L_msu(lwSum, pswSig2[i1], pswSig2[i1]);
    i1++;		/* ptr update, no count */
    test();
    if (L_sub(lwSum, lwMax) > 0)
      {
      lwMax = lwSum;			move32();
      iLoc = i1;			move16();
      }
    }


  /*
     Continue to move the window as if the processing interval
     is expanded cyclically
     i1             i2
     |     i2           --------->               |
     |-----|---------------------------|---------|   
     |                                           | 


     ... up to this position
     i1             i2
     |             i2   --------->               |
     |-------------|----------------------------||   
     |                                           | 

   */
  iLim = sub(iLim, 1);    
  for (i2=0; i2 < iLim; i2++)
    {
    lwSum = L_mac(lwSum, pswSig1[i2], pswSig1[i2]);
    lwSum = L_msu(lwSum, pswSig1[i1], pswSig1[i1]);
    lwSum = L_mac(lwSum, pswSig2[i2], pswSig2[i2]);
    lwSum = L_msu(lwSum, pswSig2[i1], pswSig2[i1]);
    i1++;		/* ptr update, no count */
    test();
    if (L_sub(lwSum, lwMax) > 0)
      {
      lwMax = lwSum;			move32();
      iLoc = i1;			move16();
      }
    }

  return iLoc;

  }


/*
***********************************************************************
*
*   FUNCTION NAME:
*
*      find_most_energetic_window_be
*
*   PURPOSE:
*
*      To find the most energetic window for the given signal
*
*   INPUTS:
*
*      *pswSig - Input signal
*
*      iLength - Pitch period length
*
*      iWinLength - Window length of the signal
*
*   OUTPUTS:
*
*      None
*
*   RETURN VALUE:
*
*      The most energetic position in the window is returned
*
**********************************************************************
*/
/* ------------------------------------------------------------------   
  Find the most energetic position of the window of given
  length (iWinLength). 
 ------------------------------------------------------------------- */
static int find_most_energetic_window_be(
                      Shortword *pswSig,                      
                      X_INT16   iLength,
                      X_INT16   iWinLength
                      )
  {
  Longword lwSum, lwMax;
  X_INT16 i1, i2, iLoc;

  lwSum = 0;		move32();


  /* --------------------------------------------------------
     Initial window location

     i1             i2
     |                                           |
     |--------------|----------------------------|   
     |                                           | 

     -------------------------------------------------------- */

  for (i2=0; i2 < iWinLength; i2++)
    {
    lwSum = L_mac(lwSum, pswSig[i2], pswSig[i2]);
    }    
  lwMax = lwSum;				move32();
  iLoc = 0;					move16();


  /*
     Move the window up to the leftmost position
     i1             i2
     |         --------->                        |
     |----------------------------|--------------|   
     |                                           | 

   */
  i1=0;			move16();
  for ( ; i2 < iLength; i2++)
    {
    lwSum = L_mac(lwSum, pswSig[i2], pswSig[i2]);
    lwSum = L_msu(lwSum, pswSig[i1], pswSig[i1]);
    i1++;		/* ptr update, no count */
    test();
    if (L_sub(lwSum, lwMax) > 0)
      {
      lwMax = lwSum;				move32();
      iLoc = i1;				move16();
      }
    }

  return iLoc;
  }


/*
***********************************************************************
*
*   FUNCTION NAME:
*
*      compute_pcorr_be
*
*   PURPOSE:
*
*      To compute the periodic correlation value of the pitch candidate
*
*   INPUTS:
*
*      pswSigIn[] - Processed, downsampled speech input
*
*      lwPitchFreq - Pitch frequency
*
*      iFrameNo - Frame number
*
*   OUTPUTS:
*
*      *pswCorr - Periodic correlation value
*
*   RETURN VALUE:
*
*      None
*
**********************************************************************
*/

void compute_pcorr_be(
                   Shortword *pswSigIn,
                   Longword lwPitchFreq,
                   X_INT16  iFrameNo,
                   Shortword *pswCorr
                   )
  {


  /*
   ********************************************************************
   *                                                                  *
   *                         Static variables                         *
   *                                                                  *
   ********************************************************************
   */

  static X_INT16 iOldPitchPeriod = -1;
  static X_INT16 iOldFrameNo = -1;
  static PCORR_STATE_be s_be;


  /*
   ********************************************************************
   *                                                                  *
   *                        Automatic variables                       *
   *                                                                  *
   ********************************************************************
   */

  X_INT16 iPitchPeriod, iWinLength;
  X_INT16 iX1Index;

  Shortword swAlpha, swBeta;
  Longword lwTmp;
  X_INT16 iTmp;
  Shortword swTmp;
  X_INT16 iIntPeriod;
  Shortword swFracPeriod;

  X_INT16 i, iStartIndex, iShift;
  Shortword swMax;
  Shortword pswSigSpace[FRAME_LEN_DS+HISTORY_LEN_DS];
  Shortword *pswSignal = pswSigSpace + HISTORY_LEN_DS; move16();


  /*
   ********************************************************************
   *                                                                  *
   *                         Executable Code                          *
   *                                                                  *
   ********************************************************************
   */

  test();
  if (lwPitchFreq == 0)
    {
    *pswCorr = 0;			move16();
    return;
    }


  /*
   * Determine the pitch period and interpolation factors
   */

  iTmp = norm_l(lwPitchFreq);
  lwTmp = L_shl(lwPitchFreq, iTmp);
  swTmp = round(lwTmp);
  swTmp = div_s(0x1F40, swTmp);	/* Q0/Q(iTmp) => Q(15-iTmp) */
  lwTmp = L_deposit_h(swTmp);
  iTmp = sub(15, iTmp);
  lwTmp = L_shr(lwTmp, iTmp);

  iIntPeriod = extract_h(lwTmp);
  lwTmp = L_mac(lwTmp, iIntPeriod, 0x8000);
  lwTmp = L_shl(lwTmp, 15);
  swFracPeriod = extract_h(lwTmp);


  iPitchPeriod = add(shr(iIntPeriod,2),1);
  swBeta = add(shl((iIntPeriod&0x0003),13),shr(swFracPeriod,2)); logic16();
  swAlpha = negate(add(0x8000,swBeta));


  /*
   * Scaling - we will make sure that the peak of the signal
   * in the appropriate range is less than 6827 - This will
   * ensure that the sum of squares over 2*WINDOW_LENGTH (36)
   * samples is less than 1 assuming a peak-to-rms factor of
   * 1.25 which is less than that of a sine wave signal.
   */

  test();
  sub(iPitchPeriod, FRAME_LEN_DS_BY_2);
  if(iPitchPeriod < FRAME_LEN_DS_BY_2)
  {
    iStartIndex = 0;				move16();
  }
  else
  {
    iStartIndex = sub(FRAME_LEN_DS,shl(iPitchPeriod,1));
  }

  test();
  add(iStartIndex, HISTORY_LEN_DS);
  if (iStartIndex < -HISTORY_LEN_DS)
  {
    iStartIndex = HISTORY_LEN_DS;		move16();
  }

  swMax = 1;					move16();
  for (i = iStartIndex; i < FRAME_LEN_DS; i++)
  {
    swTmp = abs_s(pswSigIn[i]);
    test();
    if (sub(swTmp,swMax) > 0)
    {
      swMax = swTmp;				move16();
    }
  }
         
  iShift = 0;					move16();
  test();
  if (sub(swMax,6827) > 0)
  {
    test();
    while (sub(swMax,6827) > 0)
    {
      iShift = sub(iShift,1);
      swMax = shr(swMax,1);
      test();
    }
  }
  else
  {
    test();
    while (sub(6827,swMax) >= 0)
    {
      iShift = add(iShift,1);
      swMax = shl(swMax,1);
      test();
    }
    iShift = sub(iShift,1);
  }

  for (i = -HISTORY_LEN_DS; i < FRAME_LEN_DS; i++)
  {
    pswSignal[i] = shl(pswSigIn[i],iShift);		move16();
  }

  test(); test(); sub(0,0); sub(0,0);
  if (iOldFrameNo == iFrameNo && iPitchPeriod == iOldPitchPeriod)
    {
    /*
     * If the current pitch period is the same as old pitch
     * period, use the previously calculated correlations and
     * energies for computing PCorr value
     */
    *pswCorr = interpolate_be(&s_be, swAlpha, swBeta);	move16();

    }
  else
    {

    /*
     * Determine the starting indices
     */

    test();
    if (sub(iPitchPeriod, FRAME_LEN_DS_BY_2) < 0)
      {
      iX1Index = FRAME_LEN_DS_BY_2;			move16();
      }
    else
      {
      iX1Index = sub(FRAME_LEN_DS, iPitchPeriod);
      }


    /*
     * Compute the correlations, energies, and sums 
     */

    s_be.swX1_Sum = 0;		move16();
    s_be.swZ1_Sum = 0;		move16();

    s_be.lwX1_X1 = 0;		move32();
    s_be.lwZ1_Z1 = 0;		move32();

    s_be.lwX1_Z1 = 0;		move32();
    s_be.lwX1_Z2 = 0;		move32();
    s_be.lwZ1_Z2 = 0;		move32();

    s_be.swZ2_Sum = 0;		move16();
    s_be.lwZ2_Z2 = 0;		move32();


    // We use the same window length (75 samples for 8kHz speech,
    // i.e. about 9ms) regardless of pitch period
    //         
    iWinLength = WINDOW_LENGTH;		move16();



    test(); sub(iPitchPeriod, iWinLength);
    if (iPitchPeriod <= iWinLength) {
      // The window conatins more than entire period
      iX1Index = find_most_energetic_window_be(pswSignal, FRAME_LEN_DS,
	  iPitchPeriod + iWinLength);
      iX1Index = add(iX1Index ,iPitchPeriod);

      accumulate_be(&s_be, pswSignal, iX1Index, iPitchPeriod, iWinLength);
    }
    else {
      // The window contains a part of the period.
      // Find the most energetic position for the window,
      // Take into account the cyclic nature of the signal buffer - 
      // the window can be splitted into two sub-windows
      int iCyclicWinLoc, iSubWinLength;
      int iOffset;
      iOffset = sub(iX1Index, iPitchPeriod);
      test(); sub(iOffset, -25); 
      if (iOffset < -25)
	{
	iOffset = add(iX1Index, 25);
	}
      else
	{
	iOffset = iPitchPeriod;		move16();
	}
      iCyclicWinLoc = find_most_energetic_window2_be(
	  &pswSignal[iX1Index],                                      
	  &pswSignal[iX1Index-iOffset],
	  iPitchPeriod,
	  iWinLength);
      test(); sub(add(iCyclicWinLoc, iWinLength), iPitchPeriod);
      if (iCyclicWinLoc + iWinLength <= iPitchPeriod) {
	// Take the whole window and correlate
	iX1Index = add(iX1Index, iCyclicWinLoc);

	accumulate_be(&s_be, pswSignal, iX1Index, iOffset, iWinLength);
      }
      else {
	// Split the window into two sub-windows
	iSubWinLength = sub(iPitchPeriod, iCyclicWinLoc);

	accumulate_be(&s_be, pswSignal, iX1Index + iCyclicWinLoc,
	    iOffset, iSubWinLength);
	iSubWinLength = sub(iWinLength, iSubWinLength);

	accumulate_be(&s_be, pswSignal, iX1Index, iOffset, iSubWinLength);
      }
    }

    /*
     * Adjust for DC levels
     */

    lwTmp = L_mult(s_be.swX1_Sum, s_be.swX1_Sum);
    lwTmp = Mpy_lw_sw(lwTmp, INV_WINDOW_LENGTH);
    s_be.lwX1_X1 = L_sub(s_be.lwX1_X1, lwTmp);            move32();

    lwTmp = L_mult(s_be.swZ1_Sum, s_be.swZ1_Sum);
    lwTmp = Mpy_lw_sw(lwTmp, INV_WINDOW_LENGTH);
    s_be.lwZ1_Z1 = L_sub(s_be.lwZ1_Z1, lwTmp);            move32();

    lwTmp = L_mult(s_be.swZ2_Sum, s_be.swZ2_Sum);
    lwTmp = Mpy_lw_sw(lwTmp, INV_WINDOW_LENGTH);
    s_be.lwZ2_Z2 = L_sub(s_be.lwZ2_Z2, lwTmp);            move32();

    lwTmp = L_mult(s_be.swX1_Sum, s_be.swZ1_Sum);
    lwTmp = Mpy_lw_sw(lwTmp, INV_WINDOW_LENGTH);
    s_be.lwX1_Z1 = L_sub(s_be.lwX1_Z1, lwTmp);            move32();

    lwTmp = L_mult(s_be.swX1_Sum, s_be.swZ2_Sum);
    lwTmp = Mpy_lw_sw(lwTmp, INV_WINDOW_LENGTH);
    s_be.lwX1_Z2 = L_sub(s_be.lwX1_Z2, lwTmp);            move32();

    lwTmp = L_mult(s_be.swZ1_Sum, s_be.swZ2_Sum);
    lwTmp = Mpy_lw_sw(lwTmp, INV_WINDOW_LENGTH);
    s_be.lwZ1_Z2 = L_sub(s_be.lwZ1_Z2, lwTmp);            move32();

    *pswCorr = interpolate_be(&s_be, swAlpha, swBeta);	move16();

    iOldPitchPeriod = iPitchPeriod;		move16();
    iOldFrameNo = iFrameNo;			move16();

    }

  }


