/*=======================================================================================================*/
/*      Distributed Speech Recognition                                                                   */
/*      ETSI ES 202 050 Advanced Front-End Feature Extraction Algorithm & Compression Algorithm &        */
/*      ETSI ES 202 212 Extended Advanced Front-End Feature Extraction Algorithm & Compression Algorithm */  
/*      Bit exact C-language software implementation using ETSI basic operations library                 */
/*      3GPP TS 26.243 v 6.1.0 November 2004                                                                  */
/*      ANSI C code for the Fixed-Point Distributed Speech Recognition Extended Advanced Front-end       */
/*=======================================================================================================*/


/*
**********************************************************************
*                                                                    *
*                           Include Files                            *
*                                                                    *
**********************************************************************
*/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#include "preProc_B.h"
#include "typedef.h"
#include "basic_op.h"
#include "count.h"


/*
**********************************************************************
*                                                                    *
*                                Defines                             *
*                                                                    *
**********************************************************************
*/


/*
***********************************************************************
*
*   FUNCTION NAME: pre_process
*
*   PURPOSE: To pre-process a speech frame for pitch extraction and
*            classification purposes
*
*   INPUTS:
*
*      pswInpSpeech[0:FRAME_LEN-1] - Input speech buffer
*
*      iFirstFrameFlag - Flag to indicate the first frame
*
*      iLowBandNoiseFlag - Flag to indicate noise in low band
*
*      pswScratch - Scratch memory
*
*   OUTPUTS:
*
*      pswUBSpeech[0:FRAME_LEN-1] - Upper Band speech
*
*      pswDownSampledProcSpeech[0:((HISTORY_LEN+FRAME_LEN)/DSF)-1]
*          - Down-sampled processed (low-pass filtered) speech
*
*   RETURN VALUE:
*
*      None
*
**********************************************************************
*/

  /*
   * Sixth order Butterworth high-pass filter
   * coefficients (1200 Hz cut-off frequency)
   * in 3 second order sections
   * (Format: b2, b1, b0, -a2, -a1)
   */

   static const Shortword pswHpfCoef[15] = {
      8007,  -16014,   8007,   -2010,  10812,
      5448,  -10874,   5427,   -4460,  12252,
     14896,  -29849,  14954,  -10711,  15926
   };


  /*
   * Sixth order Butterworth low-pass filter
   * coefficients (800 Hz cut-off frequency)
   * in 3 second order sections
   * (Format: b2, b1, b0, -a2, -a1)
   */

   static const Shortword pswLpfCoef[15] = {
      269,   539,   269,   -4517,  16909,
     1302,  2600,  1297,   -6763,  18727,
     4271,  8558,  4287,  -12057,  23009
   };


  /*
   * Low frequency emphasis filter to be used in conjunction
   * with the low-pass filter under normal conditions
   * B = [1.0 -0.6],  A = [1.0 -0.9]
   * (Format:  b1, b0, -a1)
   */

  static const Shortword pswLfeCoef[3] = {
    -9830,  16384,  14746
  };


  /*
   * States of the second order sections
   */

  static Shortword pswHpfXState[6]  = {0, 0, 0, 0, 0, 0};
  static Shortword pswHpfYState[12] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
  static Shortword pswLpfXState[6]  = {0, 0, 0, 0, 0, 0};
  static Shortword pswLpfYState[12] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
  static Shortword pswLfeXState[1]  = {0};
  static Shortword pswLfeYState[2]  = {0, 0};


/*
**********************************************************************
*                                                                    *
*                            The function                            *
*                                                                    *
**********************************************************************
*/

void pre_process(Shortword *pswInpSpeech,                  
                 X_INT16 iFirstFrameFlag,
                 X_INT16 iLowBandNoiseFlag,
                 Shortword *pswUBSpeech,
                 Shortword *pswDownSampledProcSpeech,
                 Shortword *pswScratch)
{



  /*
  ********************************************************************
  *                                                                  *
  *                        Automatic variables                       *
  *                                                                  *
  ********************************************************************
  */

  X_INT16 i,j;
  X_INT16 iLen;

  Shortword *pswTemp1;
  Shortword *pswTemp2;
  Shortword *pswTemp3;

  /*
   * Function declaration
   */

  void iir_d(const Shortword *, Shortword *, Shortword *, Shortword *,
             Shortword, Shortword, Shortword *);

  void iir_s(const Shortword *, Shortword *, Shortword *, Shortword *,
             Shortword, Shortword, Shortword *);


  /*
  ********************************************************************
  *                                                                  *
  *                         Executable Code                          *
  *                                                                  *
  ********************************************************************
  */


  /*
   * Pointer ops. for using scratch memory
   */

  pswTemp1 = pswScratch;
  pswTemp2 = pswTemp1 + (FRAME_LEN - FRAME_SHIFT);
  pswTemp3 = pswTemp2 + (FRAME_LEN - FRAME_SHIFT);


  /*
   * Differentiate between the first frame and the following frames
   */

  test();
  if (iFirstFrameFlag == TRUE)
  {

    iLen = (FRAME_LEN - FRAME_SHIFT);				move16();

    for (i = 0; i < iLen; i++)
    {
      pswTemp1[i] = shr(pswInpSpeech[i],1);			move16();
    }

  }
  else
  {

    iLen = FRAME_SHIFT;						move16();

    for (i = 0; i < FRAME_LEN-FRAME_SHIFT; i++)
    {
      pswUBSpeech[i] = pswUBSpeech[i+FRAME_SHIFT];		move16();
    }

    for (i = 0; i < (HISTORY_LEN+FRAME_LEN-FRAME_SHIFT)/4; i++)
    {
      pswDownSampledProcSpeech[i] = pswDownSampledProcSpeech[i+(FRAME_SHIFT/4)];
								move16();
    }

    for (i = 0; i < iLen; i++)
    {
      pswTemp1[i] = shr(pswInpSpeech[FRAME_LEN-FRAME_SHIFT+i],1);move16();
    }

  }


  /*
   * Perform HPF to obtain the Upper Band Speech
   */

  iir_d(pswHpfCoef,pswTemp1,pswHpfXState,pswHpfYState,
        iLen,1,pswTemp2);

  iir_d(pswHpfCoef+5,pswTemp2,pswHpfXState+2,pswHpfYState+4,
        iLen,1,pswTemp3);

  iir_d(pswHpfCoef+10,pswTemp3,pswHpfXState+4,pswHpfYState+8,
        iLen,1,pswUBSpeech+FRAME_LEN-iLen);

  for (i = 0; i < iLen; i++)
  {
    pswUBSpeech[FRAME_LEN-iLen+i] =
      shl(pswUBSpeech[FRAME_LEN-iLen+i],1);		move16();
  }


  /*
   * Perform LPF to get processed speech
   */

  test();
  if (iLowBandNoiseFlag == TRUE)
  {

    iir_d(pswLpfCoef,pswTemp1,pswLpfXState,pswLpfYState,
          iLen,1,pswTemp2);

    iir_d(pswLpfCoef+5,pswTemp2,pswLpfXState+2,pswLpfYState+4,
          iLen,1,pswTemp3);

    iir_d(pswLpfCoef+10,pswTemp3,pswLpfXState+4,pswLpfYState+8,
          iLen,1,pswTemp1);

  }
  else
  {

    iir_d(pswLpfCoef,pswTemp1,pswLpfXState,pswLpfYState,
          iLen,1,pswTemp2);

    iir_d(pswLpfCoef+5,pswTemp2,pswLpfXState+2,pswLpfYState+4,
          iLen,1,pswTemp3);

    iir_d(pswLpfCoef+10,pswTemp3,pswLpfXState+4,pswLpfYState+8,
          iLen,1,pswTemp2);

    iir_s(pswLfeCoef,pswTemp2,pswLfeXState,pswLfeYState,
          iLen,1,pswTemp1);

  }
    
  for (i = 0; i < iLen; i++)
  {
    pswTemp1[i] = shl(pswTemp1[i],1);				move16();
  }

  move16(); /* because j is initialized */
  for (i = 0, j = (HISTORY_LEN+FRAME_LEN-iLen)/4; i < iLen;
                                              i += DOWN_SAMP_FACTOR, j++)
  {
    pswDownSampledProcSpeech[j] = pswTemp1[i];			move16();
  }

  return;

}



/*
 * Private Functions
 */


/***************************************************************************
 *
 *    FUNCTION NAME: iir_d
 *
 *    PURPOSE:  Performs one second order iir section using double-precision
 *              feedback, single precision input and filter coefficients
 *
 *    INPUTS:
 *
 *      pswCoef[0:4]    An array of filter coefficients.
 *
 *      pswIn[0:npts-1] An array of input samples to be filtered
 *
 *      pswXstate[0:1]  An array containing x-state memory.
 *
 *      pswYstate[0:3]  An array containing y-state memory.
 *
 *      npts            Number of samples to filter (must be even)
 *
 *      shifts          Number of shifts to be made on output y(n) before
 *                      storing to y(n) states.
 *
 *    OUTPUTS:
 *
 *      pswOut[0:npts-1] Output array containing the filtered samples
 *
 *    RETURN:
 *
 *       none.
 *
 *    DESCRIPTION:
 *
 *       Transfer function implemented:
 *         (b0 + b1*z-1 + b2*z-2)/(1 + a1*z-1 + a2*z-2)
 *       data structure:
 *         Coeff array order:  b2,b1,b0,-a2,-a1
 *         Xstate array order: x(n-2),x(n-1)
 *         Ystate array order: y(n-2)MSW,y(n-2)LSW,y(n-1)MSW,y(n-1)LSW
 *
 *    REFERENCE:  Sub-clause 4.1.1 GSM Recommendation 06.20
 *
 *    KEYWORDS: filter, second-order section
 *
 *************************************************************************/

void   iir_d(const Shortword pswCoeff[], Shortword pswIn[], Shortword pswXstate[],
             Shortword pswYstate[], Shortword npts, Shortword shifts,
             Shortword pswOut[])
{

/*_________________________________________________________________________
 |                                                                         |
 |                            Automatic Variables                          |
 |_________________________________________________________________________|
*/

  X_INT16 loop_cnt;

  Shortword swTemp;

  Longword L_sumA;
  Longword L_sumB;


/*_________________________________________________________________________
 |                                                                         |
 |                              Executable Code                            |
 |_________________________________________________________________________|
*/

  /* Perform the filtering in a loop */
  /*---------------------------------*/

  for (loop_cnt = 0; loop_cnt < npts; loop_cnt += 2)
  {

    L_sumB = L_mult(pswYstate[1], pswCoeff[3]);
    L_sumB = L_mac(L_sumB, pswYstate[3], pswCoeff[4]);
    L_sumB = L_shr(L_sumB, 14);
    L_sumB = L_mac(L_sumB, pswYstate[0], pswCoeff[3]);
    L_sumB = L_mac(L_sumB, pswYstate[2], pswCoeff[4]);

    L_sumA = L_mac(L_sumB, pswCoeff[0], pswXstate[0]);
    L_sumA = L_mac(L_sumA, pswCoeff[1], pswXstate[1]);
    L_sumA = L_mac(L_sumA, pswCoeff[2], pswIn[loop_cnt]);

    L_sumA = L_shl(L_sumA, shifts);


    /* Update double precision Y state variables */
    /*-------------------------------------------*/

    pswYstate[0] = extract_h(L_sumA);		move16();
    swTemp = extract_l(L_sumA);
    swTemp = shr(swTemp, 2);
    pswYstate[1] = (0x3fff & swTemp);		logic16();move16();


    /* Round, store output sample and increment to next input sample */
    /*---------------------------------------------------------------*/

    pswOut[loop_cnt] = round(L_sumA);			move16();

    L_sumB = L_mult(pswYstate[3], pswCoeff[3]);
    L_sumB = L_mac(L_sumB, pswYstate[1], pswCoeff[4]);
    L_sumB = L_shr(L_sumB, 14);
    L_sumB = L_mac(L_sumB, pswYstate[2], pswCoeff[3]);
    L_sumB = L_mac(L_sumB, pswYstate[0], pswCoeff[4]);

    L_sumA = L_mac(L_sumB, pswCoeff[0], pswXstate[1]);
    L_sumA = L_mac(L_sumA, pswCoeff[1], pswIn[loop_cnt]);
    L_sumA = L_mac(L_sumA, pswCoeff[2], pswIn[loop_cnt+1]);

    L_sumA = L_shl(L_sumA, shifts);


    /* Update double precision Y state temporary variables */
    /*-----------------------------------------------------*/

    pswYstate[2] = extract_h(L_sumA);			move16();
    swTemp = extract_l(L_sumA);
    swTemp = shr(swTemp, 2);
    pswYstate[3] = (0x3fff & swTemp);		logic16();move16();


    /* Round, store output sample and increment to next input sample */
    /*---------------------------------------------------------------*/

    pswOut[loop_cnt + 1] = round(L_sumA);		move16();

    pswXstate[0] = pswIn[loop_cnt];			move16();
    pswXstate[1] = pswIn[loop_cnt+1];			move16();

  }

}



/***************************************************************************
 *
 *    FUNCTION NAME: iir_s
 *
 *    PURPOSE:  Performs one first order iir section using double-precision
 *              feedback, single precision input and filter coefficients
 *
 *    INPUTS:
 *
 *      pswCoef[0:2]    An array of filter coefficients.
 *
 *      pswIn[0:npts-1] An array of input samples to be filtered
 *
 *      pswXstate[0]    An array containing x-state memory.
 *
 *      pswYstate[0:1]  An array containing y-state memory.
 *
 *      npts            Number of samples to filter (must be even)
 *
 *      shifts          Number of shifts to be made on output y(n) before
 *                      storing to y(n) states.
 *
 *    OUTPUTS:
 *
 *      pswOut[0:npts-1] Output array containing the filtered samples
 *
 *    RETURN:
 *
 *       none.
 *
 *    DESCRIPTION:
 *
 *       Transfer function implemented:
 *         (b0 + b1*z-1)/(1 + a1*z-1)
 *       data structure:
 *         Coeff array order:  b1,b0,-a1
 *         Xstate array order: x(n-1)
 *         Ystate array order: y(n-1)MSW,y(n-1)LSW
 *
 *    REFERENCE:  Sub-clause 4.1.1 GSM Recommendation 06.20
 *
 *    KEYWORDS: filter, first order section
 *
 *************************************************************************/

void   iir_s(const Shortword pswCoeff[], Shortword pswIn[], Shortword pswXstate[],
             Shortword pswYstate[], Shortword npts, Shortword shifts,
             Shortword pswOut[])
{

/*_________________________________________________________________________
 |                                                                         |
 |                            Automatic Variables                          |
 |_________________________________________________________________________|
*/

  X_INT16 loop_cnt;

  Shortword swTemp;

  Longword L_sumA;
  Longword L_sumB;


/*_________________________________________________________________________
 |                                                                         |
 |                              Executable Code                            |
 |_________________________________________________________________________|
*/

  /* Perform the filtering in a loop */
  /*---------------------------------*/

  for (loop_cnt = 0; loop_cnt < npts; loop_cnt++)
  {

    L_sumB = L_mult(pswYstate[1], pswCoeff[2]);
    L_sumB = L_shr(L_sumB, 14);
    L_sumB = L_mac(L_sumB, pswYstate[0], pswCoeff[2]);

    L_sumA = L_mac(L_sumB, pswCoeff[0], pswXstate[0]);
    L_sumA = L_mac(L_sumA, pswCoeff[1], pswIn[loop_cnt]);

    L_sumA = L_shl(L_sumA, shifts);


    /* Update double precision Y state variables */
    /*-------------------------------------------*/

    pswYstate[0] = extract_h(L_sumA);		move16();
    swTemp = extract_l(L_sumA);
    swTemp = shr(swTemp, 2);
    pswYstate[1] = (0x3fff & swTemp);		logic16();move16();

    /* Round, store output sample and increment to next input sample */
    /*---------------------------------------------------------------*/

    pswOut[loop_cnt] = round(L_sumA);			move16();

    pswXstate[0] = pswIn[loop_cnt];			move16();

  }

}

